import asyncio
import logging
import os
import threading
import websockets
import queue
from relationalai import debugging
import __main__ # to get the users root file path

#------------------------------------------------------------------------------
# Logging Handler
#------------------------------------------------------------------------------

class WebSocketLoggingHandler(logging.Handler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from relationalai.analysis.mechanistic import Mechanism
        self.Mechanism = Mechanism
        self.queue: queue.Queue[str] = queue.Queue()

    def start(self, url: str):
        main_path = getattr(__main__, "__file__", None)
        program_span = debugging.span_start("program",  main=os.path.relpath(main_path) if main_path else "notebook")
        self.thread = threading.Thread(target=self.connect, args=(url, program_span))
        self.thread.start()

    def emit(self, record):
        d = record.msg
        if (isinstance(d, dict) and
            d["event"] == "span_start" and
            "task" in d["span"].attrs and
            "mech" not in d["span"].attrs):
            d["span"].attrs["mech"] = self.Mechanism(d["span"].attrs["task"])
        log_entry = self.format(record)
        self.queue.put_nowait(log_entry)

    def connect(self, url: str, span: debugging.Span):
        print(f"Connecting to {url}...")
        try:
            try:
                loop = asyncio.get_running_loop()
                loop.run_until_complete(self.connect_async(url, span))
            except RuntimeError:
                asyncio.run(self.connect_async(url, span))
        except (ConnectionRefusedError, websockets.WebSocketException, OSError, asyncio.TimeoutError):
            print(f"Failed to connect to {url}. Running with debug sink disabled.")

    async def connect_async(self, url: str, span: debugging.Span):
        async with websockets.connect(url) as ws:
            print("Connected.")
            while True:
                if not threading.main_thread().is_alive():
                    debugging.span_end(span)
                    log_entry = self.queue.get(timeout=1)
                    await ws.send(log_entry)
                    break

                try:
                    log_entry = self.queue.get(timeout=1)
                    await ws.send(log_entry)
                except queue.Empty:
                    pass
                except websockets.ConnectionClosedError:
                    break

#------------------------------------------------------------------------------
# Enable debugging
#------------------------------------------------------------------------------

already_debugging = False
def start_debugger_session(host = "0.0.0.0", port = 8080):
    global already_debugging
    if already_debugging:
        return

    ws_handler = WebSocketLoggingHandler()
    ws_handler.setFormatter(debugging.JsonFormatter())
    debugging.logger.addHandler(ws_handler)
    ws_handler.start(f"ws://{host}:{port}/ws/program")

    already_debugging = True



import asyncio
import json
from pathlib import Path
from typing import Any, Callable, Coroutine, Protocol

import aiohttp
import websockets
from aiohttp import web

import relationalai

PROJECT_ROOT = Path(relationalai.__file__).resolve().parent.parent.parent
HTTP_DIST_DIR = PROJECT_ROOT / "frontend" / "debugger" / "dist"

#------------------------------------------------------------------------------
# Program Connections
#------------------------------------------------------------------------------

connected_program = None

async def handle_program(req: web.Request):
    ws = web.WebSocketResponse()
    await ws.prepare(req)

    global connected_program
    if connected_program:
        print("Refusing to connect to new program until previous one finishes.")
        await ws.close()

    connected_program = ws
    buffered_broadcasts.clear()
    print("Program connected")
    try:
        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                await broadcast(msg.data)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                err = ws.exception()
                if err:
                    raise err
            else:
                raise Exception("Unknown message type", msg.type)
    finally:
        print("Program disconnected")
        connected_program = None

    return ws

#------------------------------------------------------------------------------
# Browser Client Connections
#------------------------------------------------------------------------------

connected_clients: set[web.WebSocketResponse] = set()

async def handle_client_message(msg: Any, ws: web.WebSocketResponse):
    try:
        parsed = json.loads(msg)
    except json.JSONDecodeError as e:
        print("Got invalid json from client:", e)
        return

    action = client_actions.get(parsed["type"])
    if not action:
        print("Got action request from client for non-existent action of type", parsed["type"])
        return

    kwargs = {k: v for k, v in parsed.items() if k != 'type'}
    res = action(ws, **kwargs)
    if res:
        if asyncio.iscoroutine(res):
            res = await res

        await ws.send_json(res)

async def handle_client(req: web.Request):
    ws = web.WebSocketResponse()
    await ws.prepare(req)
    connected_clients.add(ws)
    try:
        for msg in buffered_broadcasts:
            await ws.send_str(msg)

        async for msg in ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                await handle_client_message(msg.data, ws)
            elif msg.type == aiohttp.WSMsgType.ERROR:
                err = ws.exception()
                if err:
                    raise err
            else:
                raise Exception("Unknown message type", msg.type)

    except websockets.exceptions.ConnectionClosed as e:
        print(f"Client disconnected with reason: {e}")
    finally:
        connected_clients.remove(ws)

    return ws

#------------------------------------------------------------------------------
# Browser Client Actions
#------------------------------------------------------------------------------

ClientActionReturn = Coroutine[Any, Any, dict|None]|dict|None

class ClientAction(Protocol):
    def __call__(self, ws: web.WebSocketResponse, **kwargs: Any) -> ClientActionReturn:
        ...

client_actions: dict[str, ClientAction] = {}

def client_action(name: str|None = None):
    def decorator(fn: Callable[..., ClientActionReturn]):
        client_actions[name or fn.__name__] = fn

    return decorator

@client_action()
def ping(_):
    print("Pinged!")
    return {"type": "pong"}

#------------------------------------------------------------------------------
# Event Log
#------------------------------------------------------------------------------

buffered_broadcasts = []

async def broadcast(message):
    buffered_broadcasts.append(message)
    if connected_clients:
        tasks = [asyncio.create_task(client.send_str(message)) for client in connected_clients]
        await asyncio.wait(tasks)

#------------------------------------------------------------------------------
# Server
#------------------------------------------------------------------------------

async def serve_index(_):
    return web.FileResponse(HTTP_DIST_DIR / "index.html")

runner: web.AppRunner|None = None
async def run_server(host: str, port: int):
    global runner
    if runner:
        await runner.shutdown()

    app = web.Application()
    app.router.add_get("/ws/program", handle_program)
    app.router.add_get("/ws/client", handle_client)
    app.router.add_get("/", serve_index)
    app.router.add_static("/", HTTP_DIST_DIR, show_index=False)  # Serving static files from DIR
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, host, port)
    await site.start()

    print(f"Server started at http://{host}:{port}")

    return runner

def start_server(host: str, port: int):
    loop = asyncio.get_event_loop()
    loop.run_until_complete(run_server(host, port))
    loop.run_forever()


def stop_server():
    global runner
    if not runner:
        return

    loop = asyncio.get_running_loop()
    loop.run_until_complete(broadcast(json.dumps({"event": "debug_server_closed"})))
    loop.run_until_complete(runner.shutdown())
    runner = None



import relationalai as rai

model = rai.Model(name=globals().get("name", "test_union"), config=globals().get("config"))
Person = model.Type("Person")
Adult = model.Type("Adult")

LegalEntity = model.Type("LegalEntity")
Control = model.Type("Control")
TotalOwnership = model.Type("TotalOwnership")

with model.rule():
    Person.add(name="Joe", age=84)
    Person.add(name="Bob", age=40)
    Person.add(name="Jane", age=10)

with model.query() as select:
    p = Person()
    with model.union() as cool:
        with p.age > 80:
            cool.add("amazing", rating=1000)
        with p.age > 60:
            cool.add("rad", rating=100)
        with p.age > 18:
            cool.add("awesome", rating=10)
    z = select(p.name, cool, cool.rating)

print(z.results)


from collections import defaultdict
from typing import Any, Dict, Iterable, List, Set, cast

from .rel_emitter import Emitter

from .metamodel import Behavior, Builtins, Action, ActionType, Type, Var, Task
from . import metamodel as m
from . import compiler as c
from .clients import config
from .dsl import build

gather_vars = m.Utils.gather_vars

#--------------------------------------------------
# OrderedSet
#--------------------------------------------------

class OrderedSet:
    def __init__(self):
        self.set:Set[Var] = set()
        self.list:List[Var] = []

    def add(self, item):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)

    def update(self, items:Iterable[Any]):
        for item in items:
            self.add(item)

    def __contains__(self, item):
        return item in self.set

    def __bool__(self):
        return bool(self.set)

    def __iter__(self):
        return iter(self.list)

#--------------------------------------------------
# Flow
#--------------------------------------------------

class Flow():
    def __init__(self):
        self.tasks:List[Task] = []
        self.inlines:Set[Task] = set()
        self.task_bindings:Dict[Task, OrderedSet] = defaultdict(OrderedSet)
        self.task_deps:Dict[Task, Set[Task]] = defaultdict(set)
        self.provider_stack:List[Dict[Var, Task]] = [{}] # we start with a global scope
        self.task_var_mapping:Dict[Task, Dict[Var, Var]] = defaultdict(dict)

    def push_context(self):
        self.provider_stack.append({})
        neue = Task()
        self.tasks.append(neue)
        return neue

    def pop_context(self, task:Task, inline=False, ignore_deps=False, mappings:Dict[Var, Var]=dict()):
        if not ignore_deps:
            vs = [mappings.get(v, v) for v in gather_vars(task.items)]
            existing = self.fetch(vs, task)
            for e in existing:
                if e != task:
                    self.task_deps[task].add(e)
        if inline:
            self.inlines.add(task)
        self.provider_stack.pop()

    def assoc(self, vars:Iterable[Var], task:Task):
        for var in vars:
            self.provider_stack[-1][var] = task

    def has_var(self, var:Var):
        for provider in reversed(self.provider_stack):
            if var in provider:
                return True
        return False

    def fetch(self, vars:Iterable[Var], cur:Task):
        tasks:Set[Task] = set()
        for var in vars:
            for provider in reversed(self.provider_stack):
                found = provider.get(var)
                if found and found != cur:
                    tasks.add(found)
                    self.task_bindings[found].add(var)
                    break
        return tasks

    def mapped_bindings(self, task:Task, dep:Task, vars:Iterable[Var]|None=None):
        mapped = self.task_var_mapping.get(task)
        bindings = vars or self.task_bindings[dep]
        if mapped and len(mapped):
            return [mapped.get(b, b) for b in bindings]
        return bindings

    def finalize(self):
        final_tasks = []
        for neue in self.tasks:
            for dep in self.task_deps[neue]:
                if dep in self.tasks:
                    neue.items.insert(0, build.relation_action(ActionType.Get, dep, self.mapped_bindings(neue, dep)))
            if self.task_bindings[neue]:
                neue.parents.append(Builtins.InlineAnnotation)
                neue.items.append(build.relation_action(ActionType.Bind, neue, self.task_bindings[neue]))
            if neue not in self.inlines:
                final_tasks.append(neue)
        return final_tasks

    def reset(self):
        self.tasks.clear()
        self.provider_stack.clear()
        self.provider_stack.append({})
        self.task_bindings.clear()

#--------------------------------------------------
# Dataflow
#--------------------------------------------------

class Dataflow(c.Pass):
    def __init__(self, copying=True) -> None:
        super().__init__(copying)
        self.flow = Flow()

    def reset(self):
        super().reset()
        self.flow.reset()

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, task:Task, parent):
        self.query_flow(task)
        if not parent:
            final_tasks = self.flow.finalize()
            task.behavior = Behavior.Sequence
            task.items = [build.relation_action(ActionType.Call, task, []) for task in final_tasks]

    def query_flow(self, task:Task, inline=False, ignore_deps=False):
        flow = self.flow
        neue:Task = flow.push_context()

        for orig in task.items:
            ent:Any = orig.entity
            if orig.action == ActionType.Call and self.local_task(ent):
                flow.assoc(gather_vars(neue.items), neue)
                behavior = ent.value.behavior
                if behavior == Behavior.Union:
                    neue = self.union_call(orig, neue)
                elif behavior == Behavior.OrderedChoice:
                    neue = self.ordered_choice_call(orig, neue)
                elif behavior == Behavior.Query:
                    self.walk(orig, task)
            elif ent.isa(Builtins.Quantifier):
                neue = self.quantifier_call(orig, neue)
            elif ent.isa(Builtins.Aggregate):
                neue = self.aggregate_call(orig, neue)
            else:
                neue.items.append(orig)

        flow.pop_context(neue, inline=inline, ignore_deps=ignore_deps)
        return neue

    def local_task(self, var:Var):
        return isinstance(var.value, Task) and len(var.value.items) > 0

    #--------------------------------------------------
    # Union
    #--------------------------------------------------

    def union_call(self, call:Action, parent:Task):
        has_dep = False
        result_refs = []
        # add the final get that will be in the new continued task
        result_refs.append(build.relation_action(ActionType.Get, cast(Task, call.entity.value), call.bindings.values()))
        # run through the subtasks and add them to the flow
        for item in cast(Task, call.entity.value).items:
            neue = self.query_flow(cast(Task, item.entity.value))
            has_dep = has_dep or (parent in self.flow.task_deps[neue])
            rets = [i for i in neue.items if i.entity.value == Builtins.Return]
            if len(rets):
                rets[0].entity.value = call.entity.value
                rets[0].action = ActionType.Bind
                result_refs.append(rets[0])

        # if one of the subtasks depends on the parent, we need to cut the
        # current task to prevent cycles and carry on in a new one
        if has_dep:
            orig = parent
            self.flow.pop_context(orig)
            self.flow.assoc(gather_vars(orig.items), orig)
            parent = self.flow.push_context()

            # Update the bindings for the result refs to include the vars we depend on
            for bind in result_refs:
                neue_bindings = {}
                for i, var in enumerate(self.flow.task_bindings[orig]):
                    neue_bindings[Builtins.Relation.properties[i]] = var
                prop_len = len(neue_bindings)
                for v in bind.bindings.values():
                    neue_bindings[Builtins.Relation.properties[prop_len]] = v
                    prop_len += 1
                bind.bindings = neue_bindings

        parent.items.append(result_refs[0])
        return parent

    #--------------------------------------------------
    # Ordered choice
    #--------------------------------------------------

    def ordered_choice_call(self, call:Action, parent:Task):
        ordered_choice = cast(Task, call.entity.value)
        has_dep = False
        prevs = []
        branches = []

        # depending on whether or not the the subtasks depend on the parent,
        # we'll need to update all the references to the output so that any vars
        # we depend on are included in the result, to guarantee that the choosen
        # values join with the correct original rows
        result_refs = [build.relation_action(ActionType.Get, ordered_choice, call.bindings.values())]
        for item in ordered_choice.items:
            neue:Task = self.query_flow(cast(Task, item.entity.value))
            branches.append(neue)
            # find the return statement, we'll turn it into a bind for the overall orderd_choice
            rets = [i for i in neue.items if i.entity.value == Builtins.Return]
            ret_bindings = []
            if len(rets):
                rets[0].entity.value = neue
                result_refs.append(rets[0])
                ret_bindings = rets[0].bindings.values()
            else:
                rets.append(build.relation_action(ActionType.Bind, neue, ret_bindings))
                neue.items.append(rets[0])
                result_refs.append(rets[0])

            # add a bind for this particular task so that we can negate it in subsequent
            # branches, allowing us to create the ordering
            bind = build.relation_action(ActionType.Bind, ordered_choice, ret_bindings)
            result_refs.append(bind)
            neue.items.append(bind)
            # clear the ret bindings, since this should just return whether or not this
            # branch was successful
            rets[0].bindings.clear()
            # Negate all the previous branches to ensure we only return a value if
            # we're at our position in the order
            for prev in prevs:
                fetch = build.relation_action(ActionType.Get, prev, [])
                result_refs.append(fetch)
                prev_task = Task(items=[fetch])
                neue.items.append(build.call(Builtins.Not, [Var(value=[]), Var(value=prev_task)]))
            prevs.append(neue)

            has_dep = has_dep or (parent in self.flow.task_deps[neue])

        # If one of the branches depends on the parent task, we need to cut the parent
        # and create a new one to prevent cycles
        if has_dep:
            orig = parent
            self.flow.pop_context(orig)
            self.flow.assoc(gather_vars(orig.items), orig)
            # We also need to make sure that branches that don't currently depend on
            # the parent now do, since they also need to join correctly with the original
            # rows
            for branch in branches:
                self.flow.task_deps[branch].add(orig)
            # Update the bindings for the result refs to include the vars we depend on
            for bind in result_refs:
                neue_bindings = {}
                for i, var in enumerate(self.flow.task_bindings[orig]):
                    neue_bindings[Builtins.Relation.properties[i]] = var
                prop_len = len(neue_bindings)
                for v in bind.bindings.values():
                    neue_bindings[Builtins.Relation.properties[prop_len]] = v
                    prop_len += 1
                bind.bindings = neue_bindings
            parent = self.flow.push_context()

        parent.items.append(result_refs[0]) # result_refs[0] is the Get for the ordered_choice
        return parent

    #--------------------------------------------------
    # Quantifiers
    #--------------------------------------------------

    def quantifier_call(self, call:Action, parent:Task):
        quantifier = cast(Task, call.entity.value)
        group, task_var = [*call.bindings.values()]
        sub_task = self.query_flow(cast(Task, task_var.value))

        if isinstance(group.value, list) and len(group.value):
            raise Exception("TODO: grouped quantifiers")

        # Find any vars that this quantified task depends on
        sub_vars = gather_vars(sub_task.items)
        parent_vars = gather_vars(parent.items)
        # Find vars that we explicitly fetch in the child task that are used in the same
        # ent.attr pair in the parent task, so we can remove them as dependencies
        # this allows person.friend.name == "Joe" to not unify with an outer person.friend
        # constraint
        for item in sub_task.items:
            if item.action == ActionType.Get and item.entity in parent_vars:
                sub_vars -= set(item.bindings.values())
        shared = sub_vars & parent_vars
        shared.update([var for var in sub_vars if self.flow.has_var(var)])
        # bind those so we can use them in the quantified task
        sub_task.items.append(build.relation_action(ActionType.Bind, sub_task, shared))

        # create the quantified task, which just gets the subtask
        quantifed_task = Task()
        quantifed_task.items.append(build.relation_action(ActionType.Get, sub_task, shared))

        # add the call to the quantifier
        parent.items.append(build.call(quantifier, [group, Var(value=quantifed_task)]))
        return parent

    #--------------------------------------------------
    # Aggregates
    #--------------------------------------------------

    def aggregate_call(self, call:Action, parent:Task):
        orig = parent
        self.flow.pop_context(orig)
        self.flow.assoc(gather_vars(orig.items), orig)
        agg = cast(Task, call.entity.value)
        (args, group, ret) = [*call.bindings.values()]
        group_vars = cast(List[Var], group.value)

        # create the inner relation we'll aggregate over
        inner = self.flow.push_context()
        inner_vars = cast(List[Var], args.value)
        # to prevent shadowing errors we need to map the inner vars to new vars
        mapped = [Var(name=var.name, type=var.type) for var in inner_vars]
        self.flow.task_var_mapping[inner] = dict(zip(inner_vars, mapped))
        # vars that are in both the projection and grouping needed to be mapped in
        # the projection but made equivalent in the body so the grouping takes effect
        equivs = [(orig, neue) for (orig, neue) in self.flow.task_var_mapping[inner].items() if orig in group_vars]
        for (orig, neue) in equivs:
            inner.items.append(build.relation_action(ActionType.Call, Builtins.eq, [orig, neue]))
        # bind the mapped vars as the output of the inner relation
        inner.items.append(build.relation_action(ActionType.Bind, inner, mapped))
        self.flow.pop_context(inner, inline=True, mappings=dict(zip(mapped, inner_vars)))

        # create the outer aggregate
        outer = self.flow.push_context()
        outer_call = build.relation_action(ActionType.Call, agg, [Var(value=inner), ret])
        outer.items.append(outer_call)
        if agg.isa(Builtins.Extender):
            self.flow.task_bindings[outer].add(ret)
            for g in group_vars:
                self.flow.task_bindings[outer].add(g)
            for ix, var in enumerate(inner_vars):
                self.flow.task_bindings[outer].add(var)
                outer_call.bindings[Builtins.Relation.properties[ix+2]] = var
        else:
            for g in group_vars:
                self.flow.task_bindings[outer].add(g)
            self.flow.task_bindings[outer].add(ret)
        self.flow.pop_context(outer, ignore_deps=True)

        # Resume the flow
        parent = self.flow.push_context()
        parent.items.append(build.relation_action(ActionType.Get, outer, self.flow.task_bindings[outer]))
        return parent


#--------------------------------------------------
# Shredder
#--------------------------------------------------

class Shredder(c.Pass):
    def query(self, task: Task, parent):
        neue_actions = []
        for item in task.items:
            if item.action not in [ActionType.Call, ActionType.Construct] and not item.entity.isa(Builtins.Relation):
                ident, action = item.entity, item.action
                for type in item.types:
                    neue_actions.append(build.relation_action(action, type, [ident]))
                for prop, value in item.bindings.items():
                    neue_actions.append(build.relation_action(action, prop, [ident, value]))
            else:
                neue_actions.append(item)
        task.items = neue_actions

#--------------------------------------------------
# Splinter
#--------------------------------------------------

class Splinter(c.Pass):

    def query(self, task: Task, parent):
        effects = [i for i in task.items if i.action.is_effect()]

        grouped_effects = defaultdict(list)
        for item in effects:
            grouped_effects[(item.action, item.entity.value)].append(item)

        if len(grouped_effects) > 1:
            neue_items = []

            non_effects = [i for i in task.items if not i.action.is_effect()]
            effects_vars = gather_vars(effects)

            fetch = None
            if len(non_effects):
                fetch = self.create_fetch(non_effects, effects_vars)
                neue_items.append(fetch)

            for (k, b) in grouped_effects.items():
                neue_items.append(self.create_effect_query(b, effects_vars, fetch.entity.value if fetch else None))

            task.behavior = Behavior.Sequence
            task.items = neue_items

    #--------------------------------------------------
    # Subtask creation
    #--------------------------------------------------

    def create_fetch(self, non_effects: List[Action], effects_vars: Iterable[Var]):
        fetch = Task()
        fetch.parents.append(Builtins.InlineAnnotation)
        non_effects.append(build.relation_action(ActionType.Bind, fetch, effects_vars))
        fetch.items = non_effects
        return build.call(fetch, [])

    def create_effect_query(self, effects: List[Action], effects_vars: Iterable[Var], fetch: Any):
        neue = Task()
        if fetch:
            effects.insert(0, build.relation_action(ActionType.Get, fetch, effects_vars))
        neue.items = effects
        return build.call(neue, [])

#--------------------------------------------------
# SetCollector
#--------------------------------------------------

set_types = [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]

class SetCollector(c.Pass):
    def query(self, query: Task, parent):
        binds = [i for i in query.items if i.action in set_types]
        if len(binds) > 1:
            neue_items = []
            for item in query.items:
                if item.action not in set_types:
                    neue_items.append(item)
            neue_items.extend(self.create_raw(binds))
            query.items = neue_items

    def create_raw(self, binds: List[Action]):
        vals = [Var(value=[]) for i in range(len(binds[0].bindings))]
        vars = [Var() for v in vals]

        for bind in binds:
            for ix, var in enumerate(bind.bindings.values()):
                cast(List[Var], vals[ix].value).append(var)

        return [
            build.relation_action(ActionType.Get, Builtins.RawData, vals + vars),
            build.relation_action(binds[0].action, cast(Type, binds[0].entity.value), vars)
        ]
#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Clone(c.Pass):
    pass

class Compiler(c.Compiler):
    def __init__(self, config:config.Config):
        if config.get("compiler.use_v2", False):
            from . import rel2 as rel2
            super().__init__(Emitter(), [
                Clone(),
                rel2.FrameSet(),
            ])
        else:
            super().__init__(Emitter(), [
                Clone(),
                Dataflow(),
                Shredder(),
                Splinter(),
                SetCollector(),
            ])



import relationalai as rai
from relationalai.std import rel, aggregates
model = rai.Model(name=globals().get("name", ""), config=globals().get("config"))

v = 2
Foo = model.Type("Foo")

with model.rule():
    Foo.add(name="A")
    Foo.add(name="B")
    Foo.add(name="C")

with model.query() as select:
    f = Foo()
    z = select(rel.minimum(v, aggregates.count(f)))

print(z.results)


import relationalai as rai

from relationalai import debugging
from relationalai.clients import config as cfg
from relationalai.std import aggregates

from relationalai.clients.snowflake import Snowflake

# derived from https://relationalai.atlassian.net/browse/RAI-22876
def test_heap_snapshots(engine_config: cfg.Config):
    use_multi_valued = True
    use_value_types = True
    with debugging.span("run", idx=0, use_multi_valued=use_multi_valued, use_value_types=use_value_types):
        model = rai.Model("HeapSnapshotsBenchmark", config=engine_config)
        model._config.set("compiler.use_multi_valued", use_multi_valued)
        model._config.set("compiler.use_value_types", use_value_types)

        sf = Snowflake(model)

        Node = model.Type("Node")
        ArrayEdge = model.Type("ArrayEdge")
        ObjectEdge = model.Type("ObjectEdge")
        Edge = model.Type("Edge")

        ArrayTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__ARRAYS
        ObjectTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__OBJECTS
        NodeTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__NODES
        StringsTableEntry = sf.PYREL_HEAP_SNAPSHOTS_BENCHMARK.PUBLIC.EMPTY__STRINGS

        with model.query(tag="select_5") as select:
            out = select(5)
        print(out.results)

        with model.query(tag="count_strings") as select:
            s = StringsTableEntry()
            out = select(aggregates.count(s))
        print(out.results)

        with model.query(tag="count_string_values") as select:
            s = StringsTableEntry()
            out = select(aggregates.count(s.value))
        print(out.results)

        # Build the nodes using the string table to get string typename values
        with model.rule():
            n = NodeTableEntry()
            name = StringsTableEntry(id = n.type)
            Node.add(id=n.id, typename=name.value)

        # Build the edges using the string table to get string typename values
        with model.rule():
            o = ObjectTableEntry()
            label = StringsTableEntry(id = o.label)
            ObjectEdge.add(from_=o.from_, to=o.to_, label=label.value)
            Edge.add(from_=o.from_, to=o.to_)

        with model.rule():
            a = ArrayTableEntry()
            o = ObjectTableEntry()
            ArrayEdge.add(from_=o.from_, to=o.to_, index=a.index)
            Edge.add(from_=o.from_, to=o.to_)

        with model.query(tag="count_object_edges") as select:
            out = select(aggregates.count(ObjectEdge()))
        print(out.results)



#pyright: reportUnusedExpression=false
import relationalai as rai
from relationalai.clients.snowflake import Snowflake
from relationalai.clients import config as cfg
from relationalai.std import aggregates, rel
from relationalai.std.graphs import Graph

from relationalai import debugging

def test_tastybytes(engine_config: cfg.Config):
    # This benchmark expects the model to already be present, with the data loaded into it.
    model = rai.Model("PyRelTastyBytesBenchmark", config=engine_config)
    model._config.set("compiler.use_multi_valued", True)
    sf = Snowflake(model)
    Record = sf.TASTYBYTES.HARMONIZED.LOYALTY_ORDERS_REGION_CALIFORNIA

    Customer = model.Type("Customer")
    Truck = model.Type("Trucks")
    Transaction = model.Type("Transaction")
    RelevantConnection = model.Type("RelevantConnection")
    
    with model.query(tag="count_all") as select:
        record = Record()
        num_records = aggregates.count(record)
        select(num_records)

    # Define Customer Type
    with model.rule(dynamic=True):
        r = Record()
        Customer.add(customer_id=r.customer_id)

    # Check total number of customers
    with model.query(tag="count_customers") as select:
        record = Customer()
        num_records = aggregates.count(record)
        select(num_records)

    # Define Truck Type
    with model.rule():
        r = Record()
        Truck.add(truck_id=r.truck_id)

    # Check total number of trucks
    with model.query(tag="count_trucks") as select:
        record = Truck()
        num_records = aggregates.count(record)
        select(num_records)

    with model.rule():
        r = Record()
        Transaction.add(
            customer_id=r.customer_id,
            order_id=r.order_id,
            truck_id=r.truck_id,
            order_ts=r.order_ts,
            order_ts_seconds=r.order_ts_seconds,
            location_id=r.location_id,
        )

    with model.rule():
        t1 = Transaction()
        t2 = Transaction()

        t1.truck_id == t2.truck_id
        t1.customer_id != t2.customer_id
        rel.abs(t1.order_ts_seconds - t2.order_ts_seconds) <= 1200

        t1.connected.add(t2)

    with model.query(tag="count_connected_customers") as select:
        t = Transaction()
        num_records = aggregates.count(t.customer_id, t.order_ts, t.connected, t.connected.customer_id)
        select(num_records)

    with model.rule():
        t = Transaction()
        total_connections = aggregates.count(
            t, per=[t.customer_id, t.connected.customer_id]
        )
        total_connections > 4
        RelevantConnection.add(
            customer_1=Customer(customer_id=t.customer_id),
            customer_2=Customer(customer_id=t.connected.customer_id),
            total_connections=total_connections,
        )        

    # Get the total occurrences where pairs of customers coexisted together more than once
    with model.query(tag="count_meaningful_connected_customers") as select:
        record = RelevantConnection()
        num_records = aggregates.count(record)
        select(num_records)

    community_graph = Graph(model, undirected=True)

    # Add edges to the graph between customers / Nodes will be added automatically
    with model.rule():
        connection = RelevantConnection()
        community_graph.Edge.add(
            connection.customer_1,
            connection.customer_2,
            weight=connection.total_connections,
        )

    with model.rule():
        customer = Customer()
        community_id = community_graph.compute.louvain(customer)
        customer.set(community_id=community_id)

        community_graph.Node.add(
            customer,
            community_id=community_id,
            customer_id=customer.customer_id
        )

    with debugging.span("fetch", graph="louvain"):
        community_graph.fetch()


from collections import defaultdict
from typing import Dict, List, Set, cast

from .metamodel import Behavior, Builtins, Action, ActionType, Property, Type, Var, Task
from . import metamodel as m
from .dsl import build, is_static

gather_vars = m.Utils.gather_vars

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def add(list, item):
    if item not in list:
        list.append(item)

def flatten_vars(vars:List[Var]):
    res = []
    for var in vars:
        if not isinstance(var, Var):
            continue
        if isinstance(var.value, list):
            res.extend(flatten_vars(cast(List[Var], var.value)))
        else:
            res.append(var)
    return res

#--------------------------------------------------
# Frame
#--------------------------------------------------
frame_id = 0

class Frame():
    def __init__(self, task:Task, prev_frame:'Frame|None'=None):
        global frame_id
        self.id = frame_id
        frame_id += 1
        self.flow_graph:Dict[Var, Set[Action]] = defaultdict(set)
        self.roots = set()
        self.static = defaultdict(list)
        self.constraints = defaultdict(set)
        self.get_action:Action|None = None
        self.finalized = False
        self.strict = False
        self.multi_effect = False
        if prev_frame:
            self.merge(prev_frame)
        self.task = task

    def add_action(self, action):
        if not action.entity.isa(Builtins.Intermediate) \
            and not action.types and not action.bindings:
            return

        items = self.task.items
        for item in items:
            if item is action or item.equiv(action):
                return

        items.append(action)

    def stage(self, action:Action):
        for prop, value in action.bindings.items():
            if not prop.is_input or action.entity.isa(Builtins.Filter):
                self.add_dep(value, action)
        # for dep in action.vars(recursive=True):
        #     self.add_dep(dep, action)

    def add_dep(self, var:Var, action:Action):
        for dep_action in self.flow_graph[var]:
            if dep_action.equiv(action):
                return
        self.flow_graph[var].add(action)

    def pull_action(self, action:Action):
        for var in action.vars(recursive=True):
            self.pull_var(var)

        self.add_action(action)

    def pull_var(self, var:Var):
        vars = [var]
        actions = []
        for cur in vars:
            for dep_action in self.flow_graph[cur]:
                if self.is_constrained(cur, dep_action):
                    continue
                vals = [*dep_action.bindings.values()]
                if dep_action.entity.isa(Builtins.Intermediate):
                    if cur not in dep_action.vars():
                        prop = Builtins.Relation.properties[len(dep_action.bindings)]
                        dep_action.bindings[prop] = cur
                if isinstance(dep_action.entity.value, Property) and cur is vals[0]:
                    if vals[1].value is not None:
                        add(actions, dep_action)
                    elif self.has_other_dep(vals[1], dep_action) and not self.is_constrained(vals[1], dep_action):
                        add(actions, dep_action)
                        add(vars, vals[1])
                else:
                    add(actions, dep_action)
                    for sub_var in dep_action.vars(recursive=True):
                        add(vars, sub_var)

        for action in reversed(actions):
            self.add_action(action)
            self.constraints[var].add(action)
        return actions

    def has_other_dep(self, var:Var, action:Action):
        for dep_action in self.flow_graph[var]:
            if var in dep_action.vars() and not dep_action.equiv(action):
                return True
        return False

    def is_constrained(self, var:Var, action:Action):
        for dep_action in self.constraints[var]:
            if dep_action.equiv(action):
                return True
        return False

    #--------------------------------------------------
    # Frame management
    #--------------------------------------------------

    def merge(self, other:'Frame'):
        self.merge_constraints(other)
        # self.roots.update(other.roots)
        for var, actions in other.flow_graph.items():
            self.flow_graph[var].update(actions)

    def merge_constraints(self, other:'Frame'):
        for var, actions in other.constraints.items():
            self.constraints[var].update(actions)

#--------------------------------------------------
# FrameSet
#--------------------------------------------------

class FrameSet():
    def __init__(self):
        self.frames:List[Frame] = []
        self.stack:List[Frame] = []
        self.static = defaultdict(list)
        self.binds = {}
        self.raw = []

    #--------------------------------------------------
    # Walk
    #--------------------------------------------------

    def walk(self, item: m.AllItems, parent=None) -> m.AllItems:
        if isinstance(item, Task):
            self.task(item)
            calls = [build.call(f.task, []) for f in self.frames if f.task.items]
            return Task(behavior=Behavior.Sequence, items=self.raw + calls)
        else:
            raise Exception(f"TODO: Demand for {type(item)}")

    def task(self, task: Task, strict=False):
        return getattr(self, task.behavior.value)(task, strict)

    #--------------------------------------------------
    # Sequence
    #--------------------------------------------------

    def sequence(self, task: Task, strict):
        for item in task.items:
            if item.is_subtask_call():
                self.task(cast(Task, item.entity.value), strict)
            elif item.entity.isa(Builtins.RawCode):
                self.raw.append(item)
            else:
                raise Exception(f"Non call sequence item: {item}")

    #--------------------------------------------------
    # Union
    #--------------------------------------------------

    def union(self, task: Task, strict):
        inputs = self.get_inputs(task)
        for sub in task.items:
            if not sub.is_subtask_call():
                raise Exception(f"Non-scope in a union: {sub}")
            neue_items = []
            for item in cast(Task, sub.entity.value).items:
                if item.entity.isa(Builtins.Return):
                    item = build.relation_action(ActionType.Bind, task, [*inputs] + [*item.bindings.values()])
                neue_items.append(item)

            neue = Task(behavior=Behavior.Query, items=neue_items)
            self.task(neue)

    #--------------------------------------------------
    # Ordered Choice
    #--------------------------------------------------

    def ordered_choice(self, task: Task, strict):
        res = []
        prevs = []
        inputs = self.get_inputs(task)

        for sub in task.items:
            if not sub.is_subtask_call():
                raise Exception(f"Non-scope in an ordered_choice: {sub}")
            neue = Task(behavior=Behavior.Query)
            for item in cast(Task, sub.entity.value).items:
                if item.entity.isa(Builtins.Return):
                    item = build.relation_action(ActionType.Bind, task, [*inputs] + [*item.bindings.values()])
                neue.items.append(item)

            if prevs:
                # add a not around the previous
                for prev in prevs:
                    get = build.relation_action(ActionType.Get, prev, inputs)
                    not_task = Task(behavior=Behavior.Query, items=[get], parents=[Builtins.Inline])
                    neue.items.insert(0, build.call(Builtins.Not, [[], not_task]))

            if not sub == task.items[-1]:
                prev = build.relation("", len(inputs))
                neue.items.append(build.relation_action(ActionType.Bind, prev, inputs))
                prevs.append(prev)

            self.task(neue)
        return res

    #--------------------------------------------------
    # Shred
    #--------------------------------------------------

    def shred(self, sub:Action):
        actions = []
        ident, action = sub.entity, sub.action
        for type in sub.types:
            actions.append(build.relation_action(action, type, [ident]))
        for prop, value in sub.bindings.items():
            actions.append(build.relation_action(action, prop, [ident, value]))
        return actions

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, task:Task, strict):
        orig_items = task.items
        task.items = []
        frame = self.push_frame(task=task, strict=strict)

        # check if this has multiple effects
        effect_count = 0
        for sub in reversed(orig_items):
            if sub.is_subtask_call():
                effect_count += 1
            elif sub.entity.isa(Builtins.Return):
                effect_count += 1
            elif sub.action.is_effect() and sub.entity.isa(Builtins.Relation):
                effect_count += 1
            elif sub.action.is_effect():
                effect_count += len(sub.types) + len(sub.bindings)
            if effect_count > 1:
                break
        frame.multi_effect = effect_count > 1

        for sub in orig_items:
            if sub.action not in [ActionType.Call, ActionType.Construct] and not sub.entity.isa(Builtins.Relation):
                cur = self.frame()
                shredded = self.shred(sub)
                if sub.entity.value is None and shredded:
                    cur.roots.add(sub.entity)
                if sub.action.is_effect():
                    for shred in shredded:
                        self.pull_deps(shred)
                else:
                    for shred in shredded:
                        cur.stage(shred)
            elif sub.is_subtask_call():
                self.split(sub)
                self.task(cast(Task, sub.entity.value))
                if sub.bindings:
                    inputs = self.get_inputs(cast(Task, sub.entity.value))
                    self.frame().stage(build.relation_action(ActionType.Get, sub.entity.value, list(inputs) + [*sub.bindings.values()]))
            elif sub.entity.isa(Builtins.Install):
                self.frame().pull_action(sub)
            elif sub.entity.isa(Builtins.Quantifier):
                neue = self.quantifier(sub)
                self.frame().pull_action(neue)
            elif sub.entity.isa(Builtins.Aggregate):
                self.aggregate(sub)
            elif sub.action.is_effect():
                self.pull_deps(sub)
            else:
                self.frame().stage(sub)

        for ((root, root_action), binds) in self.static.items():
            items = []
            vals = [Var(value=[]) for i in range(len(binds[0].bindings))]
            vars = [Var() for v in vals]

            for bind in binds:
                for ix, var in enumerate(bind.bindings.values()):
                    cast(List[Var], vals[ix].value).append(var)

            # if len(effects):
            #     items.append(build.relation_action(ActionType.Get, root_out, roots))

            items.append(build.relation_action(ActionType.Get, Builtins.RawData, vals + vars))
            items.append(build.relation_action(root_action, root, vars))

            self.push_frame(Task(behavior=Behavior.Query, items=items))
            self.pop_frame()

        popped = self.pop_frame()
        while popped and popped != frame:
            popped = self.pop_frame()


    #--------------------------------------------------
    # Dep Handling
    #--------------------------------------------------

    def pull_deps(self, action:Action):
        did_split = self.split(action)

        if action.action == ActionType.Bind and self.is_static(action):
            self.static[(action.entity.value, action.action)].append(action)
            return

        self.frame().pull_action(action)

        if action.action.is_effect() and did_split:
            self.pop_frame()
        elif action.action.is_effect():
            self.frame().finalized = True

    #--------------------------------------------------
    # Quantifier
    #--------------------------------------------------

    def quantifier(self, action: Action, parent=None):
        group, task = action.bindings.values()
        if task.isa(Builtins.Inline):
            return action
        if group.value:
            raise Exception("Implement grouped quantifiers")

        task = cast(Task, task.value)
        inputs = self.get_inputs(task)

        # build a sub task that represents the body of the not, it binds
        # a relation to the inputs that we can then query for inline
        rel = build.relation("", len(inputs))
        bind = build.relation_action(ActionType.Bind, rel, inputs)
        new = Task(behavior=Behavior.Query, items=[*task.items, bind])
        self.task(new, strict=True)

        # build up an inline query for the actual quantification call
        get = build.relation_action(ActionType.Get, rel, inputs)
        inline_task = Task(behavior=Behavior.Query, items=[get], parents=[Builtins.Inline])
        cur = build.call(action.entity.value, [[], inline_task])

        return cur

    #--------------------------------------------------
    # Aggregate
    #--------------------------------------------------

    def aggregate(self, action: Action):
        # self.frame().finalized = True
        self.split(action)
        # self.pop_frame()

        projection, group, result = action.bindings.values()
        projection = cast(List[Var], projection.value).copy()
        orig_projection = projection.copy()
        group = cast(List[Var], group.value)
        agg = cast(Type, action.entity.value)
        is_extender = agg.isa(Builtins.Extender)

        proj_task = Task(behavior=Behavior.Query, parents=[Builtins.Inline])
        self.push_frame(proj_task)
        for g in group:
            if g in projection:
                new = Var()
                projection[projection.index(g)] = new
                self.frame().pull_action(build.eq(g, new))
            else:
                self.frame().pull_var(g)

        proj_bind = build.relation_action(ActionType.Bind, proj_task, [*projection])
        self.frame().pull_action(proj_bind)
        # for v in projection:
        #     self.constrain(v)
        # for v in group:
        #     self.constrain(v)
        proj_frame = self.pop_frame()
        self.frame().merge_constraints(proj_frame)

        rel = build.relation("", len(group) + 1)
        if is_extender:
            mapped_projection = [Var() for v in projection]
            call = build.relation_action(ActionType.Call, agg, [proj_task, result, *mapped_projection])
            bind = build.relation_action(ActionType.Bind, rel, [*group, result, *mapped_projection])
        else:
            call = build.relation_action(ActionType.Call, agg, [proj_task, result])
            bind = build.relation_action(ActionType.Bind, rel, [*group, result])
        t = Task(behavior=Behavior.Query, items=[call, bind])
        self.push_frame(t)
        self.pop_frame()

        if is_extender:
            get = build.relation_action(ActionType.Get, rel, [*group, result, *orig_projection])
        else:
            get = build.relation_action(ActionType.Get, rel, [*group, result])

        self.frame().stage(get)
        return get

    #--------------------------------------------------
    # Frame management
    #--------------------------------------------------

    def frame(self):
        return self.stack[-1]

    def push_frame(self, task:Task|None = None, strict=False):
        if not task:
            task = Task()
        if self.stack and not strict:
            sub = Frame(task, self.frame())
        else:
            sub = Frame(task)
        sub.strict = strict
        if sub and not sub.task.isa(Builtins.Inline):
            self.frames.append(sub)
        self.stack.append(sub)
        return sub

    def pop_frame(self):
        return self.stack.pop()

    #--------------------------------------------------
    # Root handling
    #--------------------------------------------------

    def root_like_use(self, var:Var, action:Action):
        return var in list(action.bindings.values())[:-1]

    def check_for_root(self, sub:Action, maybe_roots:Set[Var]):
        if not sub.entity.isa(Builtins.Relation) \
           or sub.entity.isa(Builtins.Infix) \
           or sub.entity.isa(Builtins.RawData):
            return
        bindings = flatten_vars(list(sub.bindings.values())[:-1])
        frame = self.frame()
        for var in bindings:
            if var.value is not None:
                continue
            deps = frame.flow_graph[var]
            if not any([self.root_like_use(var, dep) for dep in deps]):
                continue
            maybe_roots.add(var)

    def non_root(self, sub:Action, non_roots:Set[Var]):
        if sub.entity.isa(Builtins.InlineRawData):
            non_roots.update(sub.vars())

    def maybe_roots(self, task:Task, known_roots:Set[Var] = set()):
        maybe_roots = set(known_roots)
        non_roots = set()
        for sub in task.items:
            self.check_for_root(sub, maybe_roots)
            self.non_root(sub, non_roots)
        all_vars = gather_vars(task.items) - non_roots
        if len(all_vars) == 2:
            return all_vars

        return maybe_roots - non_roots

    #--------------------------------------------------
    # Splitting
    #--------------------------------------------------

    def split(self, action:Action):
        if not self.should_split(action):
            return

        frame = self.frame()
        if frame.finalized:
            self.push_frame()
            return True

        for var in frame.roots:
            frame.pull_var(var)

        task = frame.task
        if task.items:
            if frame.get_action is None:
                rel = build.relation(f"t{task.id}", 0)
                rel.parents.append(Builtins.Intermediate)
                frame.get_action = build.relation_action(ActionType.Get, rel, [])
                bind = build.relation_action(ActionType.Bind, rel, [])
                bind.bindings = frame.get_action.bindings
                frame.pull_action(bind)
            maybe_roots = self.maybe_roots(task, frame.roots)
            for root in maybe_roots:
                frame.flow_graph[root].clear()
                frame.flow_graph[root].add(frame.get_action)
            frame.finalized = True
            self.push_frame()
            return True
        return False

    def should_split(self, action:Action):
        frame = self.frame()
        if frame.finalized:
            return True
        if action.action.is_effect() and not frame.strict and frame.multi_effect:
            return True
        if action.is_subtask_call():
            return True
        if action.entity.isa(Builtins.Quantifier) \
            and not [*action.bindings.values()][1].isa(Builtins.Inline):
            return True
        if action.entity.isa(Builtins.Aggregate):
            return True
        return False

    #--------------------------------------------------
    # Helpers
    #--------------------------------------------------

    def is_static(self, action:Action):
        vars = gather_vars([action])
        return all([is_static(v) for v in vars])

    def get_inputs(self, task:Task):
        requires, provides, externals = self.requires_provides(task)
        remaining = (requires | provides) - externals
        inputs = set()
        if self.stack:
            return set(self.frame().flow_graph.keys()) & remaining
        return inputs

    def requires_provides(self, task:Task):
        seen = set()
        sub_requires = set()
        sub_provides = set()
        external = set()
        for i in task.items:
            r, p, _ = i.requires_provides(seen)
            sub_requires.update(r)
            sub_provides.update(p)
        return sub_requires - sub_provides, sub_provides, external

    def reset(self):
        self.frames.clear()
        self.stack.clear()
        self.static.clear()
        self.binds.clear()
        self.raw.clear()


from datetime import datetime, timedelta
import json
import textwrap
import time
from typing import Any, Tuple, List
import base64
from urllib.error import HTTPError

from pandas import DataFrame
import pandas as pd
import re
import decimal

from relationalai import debugging

from ..errors import Errors, RAIException
from ..rel_utils import assert_no_problems
from ..loaders.loader import emit_delete_import, import_file, list_available_resources
from .config import Config
from .types import ImportSource, ImportSourceFile
from .client import Client, ResourceProvider
from .. import dsl, rel, metamodel as m
from railib import api

#--------------------------------------------------
# Constants
#--------------------------------------------------

UNIXEPOCH = 62135683200000
MILLISECONDS_PER_DAY = 24 * 60 * 60 * 1000
TXN_FIELDS = ["id", "account_name", "state", "created_on", "finished_at", "duration", "database_name", "read_only", "engine_name", "query", "query_size", "tags", "user_agent", "response_format_version"]
TXN_REPLACE_MAP = {"database_name": "database", "engine_name": "engine", "account_name": "account", "user_agent": "agent"}
VALID_ENGINE_STATES = ["REQUESTED", "PROVISIONED", "PROVISIONING"]

#--------------------------------------------------
# Resources
#--------------------------------------------------

class Resources(ResourceProvider):
    def __init__(self, profile:str|None=None, config:Config|None=None):
        super().__init__(profile, config=config)
        self._ctx = None

    def _api_ctx(self):
        if not self._ctx:
            self._ctx = api.Context(**self.config.to_rai_config())
        return self._ctx

    def reset(self):
        self._ctx = None

    #--------------------------------------------------
    # Generic
    #--------------------------------------------------

    def get_version(self):
        raise Exception("Azure version not available")

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    def get_database(self, name:str):
        return api.get_database(self._api_ctx(), name)

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------

    def list_engines(self, state:str|None = None):
        return api.list_engines(self._api_ctx(), state)

    def get_engine(self, name:str):
        return api.get_engine(self._api_ctx(), name)

    def is_valid_engine_state(self, state:str):
        return state in VALID_ENGINE_STATES

    def create_engine(self, name:str, size:str, pool:str=""):
        with debugging.span("create_engine", name=name, size=size):
            return api.create_engine_wait(self._api_ctx(), name, size)

    def delete_engine(self, name:str):
        return api.delete_engine(self._api_ctx(), name)

    def suspend_engine(self, name:str):
        return api.suspend_engine(self._api_ctx(), name)

    def resume_engine(self, name:str):
        return api.resume_engine_wait(self._api_ctx(), name)

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    def list_graphs(self) -> List[Any]:
        with debugging.span("list_models"):
            return api.list_databases(self._api_ctx())

    def get_graph(self, name:str):
        with debugging.span("get_model", name=name):
            return api.get_database(self._api_ctx(), name)

    def create_graph(self, name: str):
        with debugging.span("create_model", name=name):
            return api.create_database(self._api_ctx(), name)

    def delete_graph(self, name:str):
        with debugging.span("delete_model", name=name):
            return api.delete_database(self._api_ctx(), name)

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        return api.list_databases(self._api_ctx())

    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        rel_code = self.create_models_code(models)
        results = self.exec_raw(database, engine, rel_code, readonly=False)
        if results.problems:
            return results.problems
        return []

    def delete_model(self, database:str, engine:str, name:str):
        return api.delete_model(self._api_ctx(), database, engine, name)

    def create_models_code(self, models:List[Tuple[str, str]]) -> str:
        lines = []
        for (name, code) in models:
            name = name.replace("\"", "\\\"")
            assert "\"\"\"\"\"\"\"" not in code, "Code literals must use fewer than 7 quotes."

            lines.append(textwrap.dedent(f"""
            def delete[:rel, :catalog, :model, "{name}"]: rel[:catalog, :model, "{name}"]
            def insert[:rel, :catalog, :model, "{name}"]: raw\"\"\"\"\"\"\"
            """) + code + "\n\"\"\"\"\"\"\"")
        rel_code = "\n\n".join(lines)
        return rel_code

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        raise Exception("Azure doesn't support exports")

    def create_export(self, database: str, engine: str, name: str, inputs: List[str], out_fields: List[str], code: str):
        raise Exception("Azure doesn't support exports")

    def delete_export(self, database: str, engine: str, name: str):
        raise Exception("Azure doesn't support exports")

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    def list_imports(self, model:str|None, id:str|None, status:str|None = None, name:str|None = None, creator:str|None = None):
        return [*list_available_resources(self, model, self.get_default_engine_name()).values()]

    def create_import_stream(self, source:ImportSource, model:str, rate = 1, options: dict|None = None):
        raise Exception("Azure doesn't support import streams")

    def create_import_snapshot(self, source:ImportSource, model:str, options: dict|None = None):
        assert isinstance(source, ImportSourceFile), "Azure integration only supports loading from files and URLs right now."
        import_file(self, model, source, **(options or {}))

    def delete_import(self, import_name: str, model:str):
        res = self.exec_raw(model, self.get_default_engine_name(), emit_delete_import(import_name), False)
        assert_no_problems(res)

    def set_imports_engine(self, engine:str):
        raise Exception("Azure doesn't support setting imports engine")

    def change_imports_status(self, status:str):
        raise Exception("Azure doesn't support import status changes")

    def change_stream_status(self, stream_id: str, model:str, suspend: bool):
        raise Exception("Azure doesn't support stream status changes")

    def get_import_stream(self, stream_id: str, model:str):
        raise Exception("Azure doesn't support get import streams")

    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    def exec_raw(self, database:str, engine:str, raw_code:str, readonly=True, raw_results=True, inputs: dict = {}):
        try:
            with debugging.span("transaction") as txn_span:
                ctx = self._api_ctx()
                with debugging.span("create"):
                    txn = api.exec_async(ctx, database, engine, raw_code, readonly=readonly, inputs=inputs)
                txn_id = txn.transaction["id"]
                txn_span["txn_id"] = txn_id

                # TODO: dedup with SDK
                rsp = api.TransactionAsyncResponse()
                txn = api.get_transaction(ctx, txn_id)
                start_time = time.time()

                def check_done():
                    with debugging.span("check_status"):
                        state = api.get_transaction(ctx, txn_id)["state"]
                        return api.is_txn_term_state(state)

                with debugging.span("wait", txn_id=txn_id):
                    api.poll_with_specified_overhead(
                        check_done,
                        overhead_rate=0.2,
                        start_time=start_time,
                    )

                # TODO: parallelize
                with debugging.span("fetch"):
                    rsp.transaction = api.get_transaction(ctx, txn_id)
                    rsp.metadata = api.get_transaction_metadata(ctx, txn_id)
                    rsp.problems = api.get_transaction_problems(ctx, txn_id)
                    with debugging.span("fetch_results"):
                        rsp.results = api.get_transaction_results(ctx, txn_id)

                return rsp
        except HTTPError as err:
            res = json.loads(err.read().decode())
            # RAI API uses a JSON payload in the body to explain why the request failed
            # This annotates the error with that to make the exception actually useful.
            if "engine not found" in res.get('message', ''):
                print("") # the SDK appears to print some stuff before the error message
                Errors.engine_not_found(self.config.get("engine", "Unknown"), res.get('message'))
            raise RAIException(f" {res.get('message', '')} {res.get('details', '')}")

    def _has_errors(self, results):
        if len(results.problems):
            for problem in results.problems:
                if problem.get('is_error') or problem.get('is_exception'):
                    return True

    def format_results(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        data_frame = DataFrame()
        if len(results.results):
            ret_cols = task.return_cols() if task else []
            for result in results.results:
                types = [t for t in result["relationId"].split("/") if t != "" and not t.startswith(":")]
                result_frame:DataFrame = result["table"].to_pandas()
                for i, col in enumerate(result_frame.columns):
                    if "UInt128" in types[i]:
                        result_frame[col] = result_frame[col].apply(lambda x: base64.b64encode(x.tobytes()).decode()[:-2])
                    elif "FixedDecimal" in types[i]:
                        decimal_info = re.search(r"FixedDecimal\{Int(\d+), (\d+)\}", types[i])
                        if decimal_info:
                            bits = int(decimal_info.group(1))
                            scale = int(decimal_info.group(2))
                            if bits == 128:
                                result_frame[col] = result_frame[col].apply(lambda x: (decimal.Decimal(str((int(x[1]) << 64) + int(x[0]))) if x[1] > 0 else decimal.Decimal(str(x[0]))) / (10 ** scale))
                            else:
                                result_frame[col] = result_frame[col].apply(lambda x: decimal.Decimal(str(x)) / (10 ** scale))
                    elif "Int128" in types[i]:
                        result_frame[col] = result_frame[col].apply(lambda x: (int(x[1]) << 64) + int(x[0]) if x[1] > 0 else x[0])
                    elif types[i] == "Dates.DateTime":
                        result_frame[col] = pd.to_datetime(result_frame[col] - UNIXEPOCH, unit="ms")
                    elif types[i] == "Dates.Date":
                        result_frame[col] = pd.to_datetime(result_frame[col] * MILLISECONDS_PER_DAY - UNIXEPOCH, unit="ms")
                if len(ret_cols) and len(result_frame.columns) == len(ret_cols):
                    result_frame.columns = ret_cols[0:len(result_frame.columns)]
                result["table"] = result_frame
                if ":output" in result["relationId"]:
                    data_frame = pd.concat([data_frame, result_frame], ignore_index=True)
        return (data_frame, results.problems)

    #--------------------------------------------------
    # Transactions
    #--------------------------------------------------

    def get_transaction(self, transaction_id):
        txn = api.get_transaction(self._api_ctx(), transaction_id)
        if not txn:
            return None
        created_on = txn.get("created_on")
        finished_at = txn.get("finished_at")
        duration = txn.get("duration")
        if duration:
            txn["duration"] = timedelta(milliseconds=duration)
        else:
            txn["duration"] = datetime.now() - datetime.fromtimestamp(created_on / 1000)
        if created_on:
            txn["created_on"] = datetime.fromtimestamp(created_on / 1000)
        if finished_at:
            txn["finished_at"] = datetime.fromtimestamp(finished_at / 1000)
        # Remap based on the fields we care about
        result = {TXN_REPLACE_MAP.get(k, k): v for k, v in txn.items() if k in TXN_FIELDS}
        return result

    def remap_fields(self, transactions):
        if not transactions:
            return []
        for transaction in transactions:
            for key in list(transaction.keys()):
                if key in TXN_REPLACE_MAP:
                    transaction[TXN_REPLACE_MAP[key]] = transaction.pop(key)
        return transactions

    def list_transactions(self, **kwargs):
        TERMINAL_STATES = ["COMPLETED", "ABORTED"]
        VALID_KEYS = ["id", "state"]

        state = kwargs.get("state")
        only_active = kwargs.get("only_active", False)
        options = {}

        # Azure sdk supports more than just VALID_KEYS as filters but for now we pass through those
        for k, v in kwargs.items():
            if k in VALID_KEYS and v is not None:
                # Only pass state if it is a valid terminal state
                if k == "state" and v.upper() in TERMINAL_STATES:
                    options[k] = v.upper()
                if k != "state":
                    options[k] = v
        # In Azure we store transactions in cosmos and consul
        # Cosmos if the state is terminal (COMPLETED or ABORTED) and Consul if the state is not (e.g. "RUNNING")
        # So we can not filter on active non terminal states via the options passed
        transactions = api.list_transactions(self._api_ctx(), **options)

        if not transactions:
            return []
        # We filter non terminal transactions here
        if only_active:
            transactions = [t for t in transactions if t["state"] in ["CREATED", "RUNNING", "PENDING"]]
        if (isinstance(state, str) and state.upper() not in TERMINAL_STATES):
            transactions = [t for t in transactions if t["state"] in [state.upper()]]
        return self.remap_fields(transactions)

    def cancel_transaction(self, transaction_id):
        return api.cancel_transaction(self._api_ctx(), transaction_id)

    def cancel_pending_transactions(self):
        # all transactions are executed synchronously against azure
        pass

#--------------------------------------------------
# Graph
#--------------------------------------------------

def Graph(name, *, profile:str|None=None, config:Config, dry_run:bool=False):
    client = Client(Resources(profile, config), rel.Compiler(config), name, dry_run=dry_run)
    client.install("pyrel_base", dsl.build.raw_task("""
        @inline
        def make_identity(x..., z):
            exists((u) | hash128({x...}, x..., u) and
            hash_value_uint128_convert(u, z))

        @inline
        def pyrel_default({F}, c, k..., v):
            F(k..., v) or (not F(k..., _) and v = c)

        @inline
        def pyrel_unwrap(x in UInt128, y): y = x

        declare __resource
    """))
    return dsl.Graph(client, name)



# pyright: reportUnusedExpression=false
import re
import json
import time
import decimal
import textwrap
from .. import std
from collections import defaultdict
from railib.api import TransactionAsyncResponse, poll_with_specified_overhead, _parse_metadata_proto
import requests
import snowflake.connector
import pyarrow as pa
import pandas as pd
from snowflake.connector import SnowflakeConnection
from .. import debugging
from typing import Any, Dict, Iterable, Tuple, List, cast
import base64
from pandas import DataFrame
from relationalai.clients.azure import MILLISECONDS_PER_DAY, UNIXEPOCH
from ..tools.cli_controls import Spinner
from ..clients.types import AvailableModel, Import, ImportSource, ImportSourceTable
from ..clients.config import Config
from ..clients.client import Client, ResourceProvider
from ..clients.util import scrub_exception
from .. import dsl, rel, metamodel as m
from ..errors import Errors, RAIException
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime

#--------------------------------------------------
# Constants
#--------------------------------------------------

USE_EXEC_ASYNC = True
VALID_POOL_STATUS = ["ACTIVE", "IDLE", "SUSPENDED"]
TXN_SQL_FIELDS = ["id", "database_name", "state", "abort_reason", "read_only", "created_by", "created_on", "finished_at", "duration"]
IMPORT_STREAM_FIELDS = ["ID", "CREATED_AT", "CREATED_BY", "STATUS", "REFERENCE_NAME", "REFERENCE_ALIAS", "FQ_OBJECT_NAME", "RAI_DATABASE",
                        "RAI_RELATION", "DATA_SYNC_STATUS", "PENDING_BATCHES_COUNT", "NEXT_BATCH_STATUS", "NEXT_BATCH_UNLOADED_TIMESTAMP",
                        "NEXT_BATCH_DETAILS", "LAST_BATCH_DETAILS", "LAST_BATCH_UNLOADED_TIMESTAMP", "CDC_STATUS"]
VALID_ENGINE_STATES = ["READY", "PENDING"]
COMPUTE_POOL_MAP = {
    "STANDARD_1": "XS",
    "CPU_X64_XS": "XS",
    "CPU_X64_S": "XS",
    "CPU_X64_M": "S",
    "HIGHMEM_X64_S": "M",
    "HIGH_MEMORY_4": "L",
    "HIGHMEM_X64_M": "XL",
    "HIGHMEM_S": "M",
    "HIGHMEM_M": "L",
    "HighMem|S": "M",
    "HighMem|M": "L",
}

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def type_to_sql(type) -> str:
    if type is str:
        return "VARCHAR"
    if type is int:
        return "NUMBER"
    if type is float:
        return "FLOAT"
    if type is bool:
        return "BOOLEAN"
    if type is dict:
        return "VARIANT"
    if type is list:
        return "ARRAY"
    if type is bytes:
        return "BINARY"
    if isinstance(type, dsl.Type):
        return "NUMBER"
    raise ValueError(f"Unknown type {type}")

def type_to_snowpark(type) -> str:
    if type is str:
        return "StringType"
    if type is int:
        return "IntegerType"
    if type is float:
        return "FloatType"
    if type is bool:
        return "BooleanType"
    if type is dict:
        return "MapType"
    if type is list:
        return "ArrayType"
    if type is bytes:
        return "BinaryType"
    if isinstance(type, dsl.Type):
        return "IntegerType"
    raise ValueError(f"Unknown type {type}")

#--------------------------------------------------
# Resources
#--------------------------------------------------

APP_NAME = "___RAI_APP___"

class Resources(ResourceProvider):
    def __init__(
        self,
        profile: str | None = None,
        config: Config | None = None,
        connection: SnowflakeConnection | None = None,
    ):
        super().__init__(profile, config=config)
        self._conn = connection
        self._pending_transactions: list[str] = []

    def _exec(self, code:str, params:List[Any]|None = None):
        if not self._conn:
            self._conn = snowflake.connector.connect(
                user=self.config.get('user'),
                password=self.config.get('password'),
                account=self.config.get('account'),
                warehouse=self.config.get('warehouse', ""),
                role=self.config.get('role', ""),
                client_store_temporary_credential=True,
                client_request_mfa_token=True,
            )
        try:
            return self._conn.cursor().execute(
                code.replace(APP_NAME, self.get_app_name()),
                params
            )
        except Exception as e:
            orig_message = str(e).lower()
            rai_app = self.config.get("rai_app_name", "")
            if re.search(f"database '{rai_app}' does not exist or not authorized.".lower(), orig_message):
                print("\n")
                Errors.snowflake_app_missing(rai_app)
            elif re.search(r"javascript execution error", orig_message):
                match = re.search(r"\"message\":\"(.*)\"", orig_message)
                if match:
                    message = match.group(1)
                    if "engine was deleted" in message or "engine not found" in message:
                        Errors.engine_not_found(self.config.get('engine', "Unknown"), message)
                    else:
                        raise RAIException(message) from None
            raise e


    def reset(self):
        self._conn = None

    #--------------------------------------------------
    # Databases
    #--------------------------------------------------

    def get_database(self, name:str):
        results = self._exec(f"SELECT * from {APP_NAME}.api.DATABASES WHERE name like '{name}';")
        if not results:
            return None
        db = results.fetchone()
        if not db:
            return None
        return {"id": db[0], "name": db[1], "created_by": db[2], "created_on": db[3], "deleted_by": db[4], "deleted_on": db[5], "state": db[6]}

    #--------------------------------------------------
    # Engines
    #--------------------------------------------------

    def list_engines(self, state: str|None = None):
        where_clause = f"WHERE STATUS = '{state}'" if state else ""
        statement = f"select NAME, SIZE, STATUS from {APP_NAME}.api.engines {where_clause}"
        results = self._exec(statement)
        if not results:
            return []
        return [{"name":name, "size":size, "state":state}
                for (name, size, state) in results.fetchall()]

    def get_engine(self, name:str):
        results = self._exec(f"select NAME, SIZE, STATUS from {APP_NAME}.api.engines WHERE NAME='{name}'")
        if not results:
            return None
        engine = results.fetchone()
        if not engine:
            return None
        return {"name": engine[0], "size": engine[1], "state": engine[2]}

    def is_valid_engine_state(self, state:str):
        return state in VALID_ENGINE_STATES

    def create_engine(self, name:str, size:str, pool:str = ""):
        if not pool:
            raise ValueError("Pool is required")
        try:
            with debugging.span("create_engine", name=name, size=size, pool=pool):
                self._exec(f"call {APP_NAME}.api.create_engine('{name}', '{pool}', '{size}');")
        except Exception as e:
            raise Exception(f"Failed to create engine {name} using pool {pool} and size {size}: {e}") from e

    def delete_engine(self, name:str):
        self._exec(f"call {APP_NAME}.api.delete_engine('{name}');")

    def suspend_engine(self, name:str):
        raise Exception("Not implemented")

    def resume_engine(self, name:str):
        raise Exception("Not implemented")

    #--------------------------------------------------
    # Graphs
    #--------------------------------------------------

    def list_graphs(self) -> List[AvailableModel]:
        with debugging.span("list_models"):
            results = self._exec(f"select NAME from {APP_NAME}.api.databases WHERE state <> 'DELETED'")
            if not results:
                return []
            return [{"name": name} for (name,) in results.fetchall()]

    def get_graph(self, name:str):
        with debugging.span("get_model", name=name):
            results = self._exec(f"""
                select ID, NAME, CREATED_ON, DELETED_ON, STATE
                from {APP_NAME}.api.databases
                where name='{name}' AND state <> 'DELETED'
            """)
            if not results:
                return None
            return results.fetchone()

    def create_graph(self, name: str):
        with debugging.span("create_model", name=name):
            self._exec(f"call {APP_NAME}.api.create_database('{name}');")

    def delete_graph(self, name:str):
        with debugging.span("delete_model", name=name):
            self._exec(f"call {APP_NAME}.api.delete_database('{name}');")

    #--------------------------------------------------
    # Models
    #--------------------------------------------------

    def list_models(self, database: str, engine: str):
        pass

    def create_models(self, database: str, engine: str, models:List[Tuple[str, str]]) -> List[Any]:
        rel_code = self.create_models_code(models)
        self.exec_raw(database, engine, rel_code, readonly=False)
        # TODO: handle SPCS errors once they're figured out
        return []

    def delete_model(self, database:str, engine:str, name:str):
        self.exec_raw(database, engine, f"def delete[:rel, :catalog, :model, \"{name}\"]: rel[:catalog, :model, \"{name}\"]", readonly=False)

    def create_models_code(self, models:List[Tuple[str, str]]) -> str:
        lines = []
        for (name, code) in models:
            name = name.replace("\"", "\\\"")
            assert "\"\"\"\"\"\"\"" not in code, "Code literals must use fewer than 7 quotes."

            lines.append(textwrap.dedent(f"""
            def delete[:rel, :catalog, :model, "{name}"]: rel[:catalog, :model, "{name}"]
            def insert[:rel, :catalog, :model, "{name}"]: raw\"\"\"\"\"\"\"
            """) + code + "\n\"\"\"\"\"\"\"")
        rel_code = "\n\n".join(lines)
        return rel_code

    #--------------------------------------------------
    # Exports
    #--------------------------------------------------

    def list_exports(self, database: str, engine: str):
        return []

    def create_export(self, database: str, engine: str, func_name: str, inputs: List[Tuple[str, str, Any]], out_fields: List[Tuple[str, Any]], code: str):
        sql_inputs = ", ".join([f"{name} {type_to_sql(type)}" for (name, var, type) in inputs])
        sql_out = ", ".join([f"{name} {type_to_sql(type)}" for (name, type) in out_fields])
        py_outs = ", ".join([f"StructField(\"{name}\", {type_to_snowpark(type)}())" for (name, type) in out_fields])
        py_inputs = ", ".join([name for (name, *rest) in inputs])
        safe_rel = code.replace("'", "\\'").replace("{", "{{").replace("}", "}}").strip()
        clean_inputs = []
        for (name, var, type) in inputs:
            if type is str:
                clean_inputs.append(f"{name} = {name}.replace(\"'\", \"\\'\")")
            # Replace `var` with `name` and keep the following non-word character unchanged
            pattern = re.compile(re.escape(var) + r'(\W)')
            safe_rel = re.sub(pattern, rf"{{{name}}}\1", safe_rel)
        safe_rel = safe_rel.replace("\n", "\\n").replace("\"", "\\\"")
        if py_inputs:
            py_inputs = f", {py_inputs}"
        clean_inputs = ("\n" + " "*20).join(clean_inputs)
        sql_code = textwrap.dedent(f"""
                CREATE OR REPLACE PROCEDURE {func_name}({sql_inputs})
                RETURNS TABLE({sql_out})
                LANGUAGE PYTHON
                RUNTIME_VERSION = '3.8'
                PACKAGES = ('snowflake-snowpark-python')
                HANDLER = 'handle'
                AS
                $$
                import json
                from snowflake.snowpark import DataFrame
                from snowflake.snowpark.functions import col
                from snowflake.snowpark.types import StringType, IntegerType, StructField, StructType, FloatType, MapType, ArrayType, BooleanType, BinaryType
                def handle(session{py_inputs}):
                    {clean_inputs}
                    rel_code = f"{safe_rel}"
                    results = session.sql(f"select {APP_NAME}.api.exec('{database}','{engine}','{{rel_code}}', true);")
                    table_results = []
                    for row in results.collect():
                        parsed = json.loads(row[0])
                        table_results.extend(parsed["data"])

                    try:
                        schema = StructType([{py_outs}])
                        return session.create_dataframe(table_results, schema=schema)
                    except Exception as err:
                        from collections import defaultdict
                        import textwrap
                        problems = defaultdict(dict)
                        for row in table_results:
                            if row[0] == "catalog" and row[1] == "diagnostic":
                                if row[2] in ["code", "message", "report"]:
                                    problems[row[3]][row[2]] = row[4]

                        if (len(problems)):
                            raise Exception("Failed to run exported query '{func_name}({", ".join([name for (name, *_) in inputs])})':\\n" + textwrap.indent("\\n".join([problem["code"] + " " + problem["message"] + "\\n" + problem["report"] for problem in problems.values()]), "\t")) from None
                        else:
                            raise err
                $$;""")
        start = time.perf_counter()
        self._exec(sql_code)
        debugging.time("export", time.perf_counter() - start, DataFrame(), code=sql_code)
        return

    def delete_export(self, model: str, engine: str, export: str):
        pass

    #--------------------------------------------------
    # Imports
    #--------------------------------------------------

    def imports_to_dicts(self, imports):
        if not imports:
            return []
        map = {"database_name": "database"}
        mapped = [map.get(f, f.lower()) for f in IMPORT_STREAM_FIELDS]
        return [dict(zip(mapped, row)) for row in imports]

    def change_stream_status(self, stream_id: str, model:str, suspend: bool):
        if stream_id and model:
            if suspend:
                self._exec(f"CALL {APP_NAME}.api.suspend_data_stream('{stream_id}', '{model}');")
            else:
                self._exec(f"CALL {APP_NAME}.api.resume_data_stream('{stream_id}', '{model}');")

    def change_imports_status(self, suspend: bool):
        if suspend:
            self._exec(f"CALL {APP_NAME}.api.suspend_cdc();")
        else:
            self._exec(f"CALL {APP_NAME}.api.resume_cdc();")

    def get_imports_status(self):
        results = self._exec(f"CALL {APP_NAME}.api.cdc_status();")
        if results:
            for result in results.fetchall():
                (engine, status, info) = result
                return {"engine": engine, "status": status, "info": info}
        return {}

    def set_imports_engine(self, engine:str):
        try:
            self._exec(f"CALL {APP_NAME}.api.setup_cdc('{engine}');")
        except Exception as e:
            if "engine not found" in str(e):
                raise ValueError(f"Engine {engine} not found. Please create the engine first.") from e

    def list_imports(
        self,
        id:str|None = None,
        name:str|None = None,
        model:str|None = None,
        status:str|None = None,
        creator:str|None = None,
    ) -> list[Import]:
        where = []
        if id and isinstance(id, str):
            where.append(f"LOWER(ID) = '{id.lower()}'")
        if name and isinstance(name, str):
            where.append(f"LOWER(FQ_OBJECT_NAME) = '{name.lower()}'")
        if model and isinstance(model, str):
            where.append(f"LOWER(RAI_DATABASE) = '{model.lower()}'")
        if creator and isinstance(creator, str):
            where.append(f"LOWER(CREATED_BY) = '{creator.lower()}'")
        if status and isinstance(status, str):
            where.append(f"LOWER(batch_status) = '{status.lower()}'")
        where_clause = " AND ".join(where)

        # this is roughly copied from the native app code because we don't have a way to
        # get the status of multiple streams at once and doing them individually is way
        # too slow
        statement = f"""
        select ID, RAI_DATABASE, FQ_OBJECT_NAME, CREATED_AT, CREATED_BY, nextBatch.status as batch_status, processing_errors, nextBatch.batches,
            from {APP_NAME}.api.data_streams as ds
            left JOIN (
            select
                data_stream_id,
                min_by(status, unloaded) as status,
                min_by(batch_details, unloaded) as batch_details,
                min(processing_details:processingErrors) as processing_errors,
                min(unloaded) as unloaded,
                count(*) as batches
            from {APP_NAME}.api.data_stream_batches
            group by data_stream_id
            ) nextBatch
            ON ds.id = nextBatch.data_stream_id
            {f"where {where_clause}" if where_clause else ""};
        """

        results = self._exec(statement)
        items = []
        if results:
            for stream in results.fetchall():
                (id, db, name, created_at, created_by, status, processing_errors, batches) = stream
                if status and isinstance(status, str):
                    status = status.upper()
                if processing_errors:
                    if status in ["QUARANTINED", "PENDING"]:
                        start = processing_errors.rfind("Error")
                        if start != -1:
                            processing_errors = processing_errors[start:-1]
                    else:
                        processing_errors = None
                items.append(cast(Import, {
                    "id": id,
                    "model": db,
                    "name": name,
                    "created": created_at,
                    "creator": created_by,
                    "status": status.upper() if status else None,
                    "errors": processing_errors if processing_errors != "[]" else None,
                    "batches": f"{batches}" if batches else "",
                }))
        return items

    def get_import_stream(self, stream_id:str, model:str):
        results = self._exec(f"CALL {APP_NAME}.api.get_data_stream('{stream_id}', '{model}');")
        if not results:
            return None
        return self.imports_to_dicts(results.fetchall())

    def create_import_stream(self, source:ImportSource, model:str, rate = 1, options: dict|None = None):
        assert isinstance(source, ImportSourceTable), "Snowflake integration only supports loading from SF Tables right now. Try loading your data as a table via the snowflake interface first."
        object = source.fqn

        if object.lower() in [x["name"].lower() for x in self.list_imports(model=model)]:
            return

        info = self._exec(f"SHOW OBJECTS like %s in {source.database}.{source.schema}", [source.table])
        if not info:
            raise ValueError(f"Object {source.table} not found in {source.database}.{source.schema}")
        else:
            data = info.fetchone()
            if not data:
                raise ValueError(f"Object {source.table} not found in {source.database}.{source.schema}")
            # (time, name, db_name, schema_name, kind, *rest)
            kind = data[4]

        try:
            self._exec(f"ALTER {kind} {object} SET CHANGE_TRACKING = TRUE;")
        except Exception:
            pass

        command = f"""call {APP_NAME}.api.create_data_stream(
            {APP_NAME}.api.object_reference('{kind}', '{object}'),
            '{model}',
            '{object.replace('.', '_').lower()}');"""
        try:
            self._exec(command)
        except Exception as e:
            if "ensure that CHANGE_TRACKING is enabled on the source object" in str(e):
                print("\n")
                Errors.snowflake_change_tracking_not_enabled(object, f"ALTER TABLE {object} SET CHANGE_TRACKING = TRUE;")
            raise e

        return

    def create_import_snapshot(self, source:ImportSource, model:str, _: dict|None = None):
        raise Exception("Snowflake integration doesn't support snapshot imports yet")

    def delete_import(self, import_name:str, model:str):
        self._exec(f"""call {APP_NAME}.api.delete_data_stream(
            '{import_name}',
            '{model}'
        );""")
        return

    #--------------------------------------------------
    # Exec Sync
    #--------------------------------------------------

    def _exec_sync_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs = {}):
        if inputs:
            raise Exception("Inputs aren't currently supported using exec_sync in SPCS.")

        return self._exec(f"select {APP_NAME}.api.exec('{database}','{engine}','{raw_code}', {readonly});")

    def _format_results_sync(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        parsed_results = []
        parsed_problems = []
        if results:
            for row in results:
                parsed = json.loads(row[0])
                if parsed.get("problems"):
                    for problem in parsed["problems"]:
                        parsed_problems.append(problem)
                else:
                    parsed_results.extend(parsed["data"])
        try:
            ret_cols = task.return_cols() if task else []
            data_frame = DataFrame(parsed_results, columns=ret_cols)
        except Exception:
            data_frame = DataFrame(parsed_results)
        return (data_frame, parsed_problems)

    #--------------------------------------------------
    # Exec Async
    #--------------------------------------------------

    def _check_exec_async_status(self, txn_id: str):
        """Check whether the given transaction has completed."""
        with debugging.span("check_status"):
            response = self._exec(f"CALL {APP_NAME}.api.get_transaction('{txn_id}');")
            assert response, f"No results from get_transaction('{txn_id}')"
        status: str = next(iter(response))[2]

        if status == "ABORTED":
            raise Exception(f"Transaction aborted while waiting for results '{txn_id}'")

        return status == "COMPLETED"

    def _list_exec_async_artifacts(self, txn_id: str) -> Dict[str, str]:
        """Grab the list of artifacts produced in the transaction and the URLs to retrieve their contents."""
        with debugging.span("list_results"):
            response = self._exec(f"CALL {APP_NAME}.api.list_transaction_outputs('{txn_id}');")
            assert response, f"No results from list_transaction_outputs('{txn_id}')"
            return {name: url for name, url in response}

    def _fetch_exec_async_artifacts(
        self, artifact_urls: Dict[str, str]
    ) -> Dict[str, Any]:
        """Grab the contents of the given artifacts from SF in parallel using threads."""
        contents = {}

        with requests.Session() as session:
            # Define _fetch_data inside the session context to close over 'session'
            def _fetch_data(name_url):
                name, url = name_url
                try:
                    response = session.get(url)
                    response.raise_for_status()  # Throw if something goes wrong
                    if name.endswith(".json"):
                        return name, response.json()
                    else:
                        return name, response.content
                except requests.RequestException as e:
                    raise scrub_exception(e)

            # Create a list of tuples for the map function
            name_url_pairs = list(artifact_urls.items())

            # Create ThreadPoolExecutor to handle requests in parallel
            with ThreadPoolExecutor(max_workers=5) as executor:
                # Map fetch across all URLs using the executor
                results = executor.map(_fetch_data, name_url_pairs)

                # Populate contents dictionary
                for name, data in results:
                    contents[name] = data

        return contents

    def _parse_exec_async_results(self, arrow_files: List[Tuple[str, bytes]]):
        """Mimics the logic in _parse_arrow_results of railib/api.py#L303 without requiring a wrapping multipart form."""
        results = []

        for file_name, file_content in arrow_files:
            with pa.ipc.open_stream(file_content) as reader:
                schema = reader.schema
                batches = [batch for batch in reader]
                table = pa.Table.from_batches(batches=batches, schema=schema)
                results.append({"relationId": file_name, "table": table})

        return results

    def _download_results(self, artifact_urls, txn_id):
        with debugging.span("download_results"):
            # Fetch artifacts
            artifacts = self._fetch_exec_async_artifacts(artifact_urls)

            # Parse metadata from the artifacts
            meta = _parse_metadata_proto(artifacts["metadata.proto"])
            meta_json = artifacts["metadata.json"]

            # Use the metadata to map arrow files to the relations they contain
            arrow_files_to_relations = {}
            for ix, relation in enumerate(meta.relations):
                arrow_file = relation.file_name
                relation_id = meta_json[ix]["relationId"]
                arrow_files_to_relations[arrow_file] = relation_id

            # Hydrate the arrow files into tables
            results = self._parse_exec_async_results([
                (arrow_files_to_relations[name], content) for name, content in artifacts.items() if name.endswith(".arrow")
            ])

            # Create and return the response
            rsp = TransactionAsyncResponse()
            rsp.transaction = {"id": txn_id, "state": "COMPLETED", "response_format_version": None}
            rsp.metadata = meta
            rsp.problems = artifacts.get("problems.json")  # Safely access possible missing keys
            rsp.results = results
            return rsp


    def _exec_async_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs={}):
        with debugging.span("transaction") as txn_span:
            if inputs:
                raise Exception("Inputs aren't currently supported using exec_async in SPCS.")

            with debugging.span("create"):
                response = self._exec(f"CALL {APP_NAME}.api.exec_async('{database}','{engine}', %s, {readonly});", raw_code)
                if not response:
                    raise Exception("No results from exec_async")

            # Grab the txn_id from the response
            txn_id, status = next(iter(response))
            txn_span["txn_id"] = txn_id
            # Wait for completion or failure
            if status != "COMPLETED":
                self._pending_transactions.append(txn_id)
                with debugging.span("wait", txn_id=txn_id):
                    poll_with_specified_overhead(lambda: self._check_exec_async_status(txn_id), 0.2)

            self._pending_transactions.remove(txn_id)

            with debugging.span("fetch"):
                # List the result artifacts (and the URLs to retrieve them)
                artifact_urls = self._list_exec_async_artifacts(txn_id)

                return self._download_results(artifact_urls, txn_id)

    # Copied directly from azure.py
    def _has_errors(self, results):
        if len(results.problems):
            for problem in results.problems:
                if problem['is_error'] or problem['is_exception']:
                    return True

    # Copied directly from azure.py
    def _format_results_async(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        data_frame = DataFrame()
        if len(results.results):
            ret_cols = task.return_cols() if task else []
            for result in results.results:
                types = [t for t in result["relationId"].split("/") if t != "" and not t.startswith(":")]
                result_frame:DataFrame = result["table"].to_pandas()
                for i, col in enumerate(result_frame.columns):
                    if "UInt128" in types[i]:
                        result_frame[col] = result_frame[col].apply(lambda x: base64.b64encode(x.tobytes()).decode()[:-2])
                    elif "FixedDecimal" in types[i]:
                        decimal_info = re.search(r"FixedDecimal\{Int(\d+), (\d+)\}", types[i])
                        if decimal_info:
                            bits = int(decimal_info.group(1))
                            scale = int(decimal_info.group(2))
                            if bits == 128:
                                result_frame[col] = result_frame[col].apply(lambda x: (decimal.Decimal(str((int(x[1]) << 64) + int(x[0]))) if x[1] > 0 else decimal.Decimal(str(x[0]))) / (10 ** scale))
                            else:
                                result_frame[col] = result_frame[col].apply(lambda x: decimal.Decimal(str(x)) / (10 ** scale))
                    elif "Int128" in types[i]:
                        result_frame[col] = result_frame[col].apply(lambda x: (int(x[1]) << 64) + int(x[0]) if x[1] > 0 else x[0])
                    elif types[i] == "Dates.DateTime":
                        result_frame[col] = pd.to_datetime(result_frame[col] - UNIXEPOCH, unit="ms")
                    elif types[i] == "Dates.Date":
                        result_frame[col] = pd.to_datetime(result_frame[col] * MILLISECONDS_PER_DAY - UNIXEPOCH, unit="ms")
                if len(ret_cols) and len(result_frame.columns) == len(ret_cols):
                    result_frame.columns = ret_cols[0:len(result_frame.columns)]
                result["table"] = result_frame
                if ":output" in result["relationId"]:
                    data_frame = pd.concat([data_frame, result_frame], ignore_index=True)
        return (data_frame, results.problems)

    #--------------------------------------------------
    # Exec
    #--------------------------------------------------

    def exec_raw(self, database:str, engine:str, raw_code:str, readonly=True, inputs: dict = {}):
        raw_code = raw_code.replace("'", "\\'") # @NOTE: If collapsing to a single exec, make sure to copy this line into it.
        if USE_EXEC_ASYNC:
            return self._exec_async_raw(database, engine, raw_code, readonly, inputs=inputs)
        else:
            return self._exec_sync_raw(database, engine, raw_code, readonly, inputs=inputs)

    def format_results(self, results, task:m.Task|None=None) -> Tuple[DataFrame, List[Any]]:
        if USE_EXEC_ASYNC:
            return self._format_results_async(results, task)
        else:
            return self._format_results_sync(results, task)

    #--------------------------------------------------
    # Transactions
    #--------------------------------------------------
    def txns_to_dicts(self, transactions):
        if not transactions:
            return []
        map = {"database_name": "database"}
        mapped = [map.get(f, f) for f in TXN_SQL_FIELDS]
        return [dict(zip(mapped, row)) for row in transactions]

    def get_transaction(self, transaction_id):
        results = self._exec(f"CALL {APP_NAME}.api.get_transaction(%s);", [transaction_id])
        if not results:
            return None
        txn = self.txns_to_dicts(results.fetchall())[0]
        created_on = txn.get("created_on")
        finished_at = txn.get("finished_at")
        if created_on:
            if finished_at:
                txn['duration'] = finished_at - created_on
            else:
                tz_info = created_on.tzinfo
                txn['duration'] = datetime.now(tz_info) - created_on
        return txn

    def list_transactions(self, **kwargs):
        id = kwargs.get("id", None)
        state = kwargs.get("state", None)
        limit = kwargs.get("limit", 100)
        only_active = kwargs.get("only_active", False)
        where_clause_arr = []

        if id:
            where_clause_arr.append(f"id = '{id}'")
        if state:
            where_clause_arr.append(f"state = '{state.upper()}'")
        else:
            if only_active:
                where_clause_arr.append("state in ('CREATED', 'RUNNING', 'PENDING')")

        if len(where_clause_arr):
            where_clause = f'WHERE {" AND ".join(where_clause_arr)}'
        else:
            where_clause = ""

        sql_fields = ", ".join([f.upper() for f in TXN_SQL_FIELDS])
        results = self._exec(f"select {sql_fields} from {APP_NAME}.api.transactions {where_clause} LIMIT %s", [limit])
        if not results:
            return []
        return self.txns_to_dicts(results.fetchall())

    def cancel_transaction(self, transaction_id):
        self._exec(f"CALL {APP_NAME}.api.CANCEL_TRANSACTION(%s);", [transaction_id])
        if transaction_id in self._pending_transactions:
            self._pending_transactions.remove(transaction_id)

    def cancel_pending_transactions(self):
        for txn_id in self._pending_transactions:
            self.cancel_transaction(txn_id)

    #--------------------------------------------------
    # Snowflake specific
    #--------------------------------------------------

    def get_version(self):
        results = self._exec(f"SELECT {APP_NAME}.app.get_release()")
        if not results:
            return None
        return results.fetchone()[0]

    def list_warehouses(self):
        results = self._exec("SHOW WAREHOUSES")
        if not results:
            return []
        return [{"name":name}
                for (name, *rest) in results.fetchall()]

    def list_compute_pools(self):
        results = self._exec("SHOW COMPUTE POOLS")
        if not results:
            return []
        return [{"name":name, "status":status, "min_nodes":min_nodes, "max_nodes":max_nodes, "instance_family":instance_family}
                for (name, status, min_nodes, max_nodes, instance_family, *rest) in results.fetchall()]

    def list_valid_compute_pools_by_engine_size(self, engine_size:str):
        valid_pools = self.list_compute_pools()
        if not valid_pools:
            return []
        engine_pools = [item['name'] for item in valid_pools if item['status'] in VALID_POOL_STATUS and COMPUTE_POOL_MAP.get(item['instance_family']) == engine_size]
        if not engine_pools:
            return []
        return engine_pools

    def list_roles(self):
        results = self._exec("SELECT CURRENT_AVAILABLE_ROLES()")
        if not results:
            return []
        # the response is a single row with a single column containing
        # a stringified JSON array of role names:
        row = results.fetchone()
        if not row:
            return []
        return [{"name": name} for name in json.loads(row[0])]

    def list_apps(self):
        results = self._exec("SHOW APPLICATIONS")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_databases(self):
        results = self._exec("SHOW DATABASES")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_sf_schemas(self, database:str):
        results = self._exec(f"SHOW SCHEMAS IN {database}")
        if not results:
            return []
        return [{"name":name}
                for (time, name, *rest) in results.fetchall()]

    def list_tables(self, database:str, schema:str):
        results = self._exec(f"SHOW OBJECTS IN {database}.{schema}")
        items = []
        if results:
            for (time, name, db_name, schema_name, kind, *rest) in results.fetchall():
                items.append({"name":name, "kind":kind.lower()})
        return items

    def schema_info(self, database:str, schema:str, tables:Iterable[str]):
        pks = self._exec(f"SHOW PRIMARY KEYS IN SCHEMA {database}.{schema};")
        fks = self._exec(f"SHOW IMPORTED KEYS IN SCHEMA {database}.{schema};")
        tables = ", ".join([f"'{x.upper()}'" for x in tables])
        columns = self._exec(f"""
            select TABLE_NAME, COLUMN_NAME, DATA_TYPE
            from {database.upper()}.INFORMATION_SCHEMA.COLUMNS
            where TABLE_SCHEMA = '{schema.upper()}'
            and TABLE_NAME in ({tables})
            and TABLE_CATALOG = '{database.upper()}';
        """)
        results = defaultdict(lambda: {"pks": [], "fks": {}, "columns": {}})
        if pks:
            for row in pks:
                results[row[3].lower()]["pks"].append(row[4].lower()) # type: ignore
        if fks:
            for row in fks:
                results[row[7].lower()]["fks"][row[8].lower()] = row[3].lower()
        if columns:
            for row in columns:
                results[row[0].lower()]["columns"][row[1].lower()] = row[2].lower()
        return results

#--------------------------------------------------
# Snowflake Wrapper
#--------------------------------------------------

class PrimaryKey:
    pass

class Snowflake:
    def __init__(self, model, auto_import=False):
        self._model = model
        self._auto_import = auto_import
        if not isinstance(model._client.resources, Resources):
            raise ValueError("Snowflake model must be used with a snowflake config")
        self._dbs = {}
        imports = model._client.resources.list_imports(model=model.name)
        self._import_structure(imports)

    def _import_structure(self, imports: list[Import]):
        tree = self._dbs
        # pre-create existing imports
        schemas = set()
        for item in imports:
            database_name, schema_name, table_name = item["name"].lower().split('.')
            database = getattr(self, database_name)
            schema = getattr(database, schema_name)
            schemas.add(schema)
            schema._add(table_name, is_imported=True)
        for schema in schemas:
            schema._fetch_info()
        return tree

    def __getattribute__(self, __name: str) -> 'SnowflakeDB':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._dbs:
            return self._dbs[__name]
        self._dbs[__name] = SnowflakeDB(self, __name)
        return self._dbs[__name]

class SnowflakeDB:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._schemas = {}

    def __getattribute__(self, __name: str) -> 'SnowflakeSchema':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        __name = __name.lower()
        if __name in self._schemas:
            return self._schemas[__name]
        self._schemas[__name] = SnowflakeSchema(self, __name)
        return self._schemas[__name]

class SnowflakeSchema:
    def __init__(self, parent, name):
        self._name = name
        self._parent = parent
        self._model = parent._model
        self._tables = {}
        self._imported = set()
        self._table_info = defaultdict(lambda: {"pks": [], "fks": {}, "columns": {}})

    def _fetch_info(self):
        self._table_info = self._model._client.resources.schema_info(self._parent._name, self._name, self._imported)

    def _add(self, name, is_imported=False):
        name = name.lower()
        if name in self._tables:
            return self._tables[name]
        if is_imported:
            self._imported.add(name)
        else:
            self._tables[name] = SnowflakeTable(self, name)
        return self._tables.get(name)

    def __getattribute__(self, __name: str) -> 'SnowflakeTable':
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        table = self._add(__name)
        table._lazy_init()
        return table

class SnowflakeTable(dsl.Type):
    def __init__(self, parent, name):
        super().__init__(parent._model, f"sf_{name}", ["namespace", "fqname", "describe"])
        self._name = name
        self._model = parent._model
        self._parent = parent
        self._aliases = {}
        self._finalzed = False

    def _lazy_init(self):
        parent = self._parent
        name = self._name
        if name not in parent._imported:
            if self._parent._parent._parent._auto_import:
                with Spinner(f"Creating stream for {self.fqname()}", f"Stream created for {self.fqname()}"):
                    db_name = parent._parent._name
                    schema_name = parent._name
                    self._model._client.resources.create_import_stream(ImportSourceTable(db_name, schema_name, name), self._model.name)
                print("")
                parent._imported.add(name)
            else:
                imports = self._model._client.resources.list_imports(model=self._model.name)
                for item in imports:
                    cur_name = item["name"].lower().split(".")[-1]
                    parent._imported.add(cur_name)
            if name not in parent._imported:
                Errors.snowflake_import_missing(debugging.capture_code_info(), self.fqname(), self._model.name)

        if name not in parent._table_info:
            parent._fetch_info()

        self._finalize()

    def _finalize(self):
        if self._finalzed:
            return

        self._finalzed = True
        self._schema = self._parent._table_info[self._name]
        relation_name = self.fqname().replace(".", "_").lower()
        model:dsl.Graph = self._model
        model.install_raw(f"declare {relation_name}")

        with model.rule(dynamic=True):
            prop, id, val = std.Vars(3)
            if self._schema["pks"]:
                getattr(getattr(std.rel, relation_name), self._schema["pks"][0].upper())(id, val)
            else:
                getattr(std.rel, relation_name)(prop, id, val)
            self.add(snowflake_id=id)

        for prop, prop_type in self._schema["columns"].items():
            with model.rule(dynamic=True):
                id, val = std.Vars(2)
                getattr(getattr(std.rel, relation_name), prop.upper())(id, val)
                if getattr(self, prop.lower()).is_multi_valued:
                    inst = self(snowflake_id=id)
                    getattr(inst, prop.lower()).add(val)
                else:
                    self(snowflake_id=id).set(**{prop.lower(): val})

    def namespace(self):
        return f"{self._parent._parent._name}.{self._parent._name}"

    def fqname(self):
        return f"{self.namespace()}.{self._name}"

    def describe(self, **kwargs):
        model = self._model
        for k, v in kwargs.items():
            if v is PrimaryKey:
                self._schema["pks"] = [k]
            elif isinstance(v, tuple):
                (table, name) = v
                if isinstance(table, SnowflakeTable):
                    fk_table = table
                    pk = fk_table._schema["pks"]
                    with model.rule():
                        inst = fk_table()
                        me = self()
                        getattr(inst, pk[0]) == getattr(me, k)
                        if getattr(self, name).is_multi_valued:
                            getattr(me, name).add(inst)
                        else:
                            me.set(**{name: inst})
                else:
                    raise ValueError(f"Invalid foreign key {v}")
            else:
                raise ValueError(f"Invalid column {k}={v}")
        return self

#--------------------------------------------------
# Graph
#--------------------------------------------------

def Graph(
    name,
    *,
    profile: str | None = None,
    config: Config,
    dry_run: bool = False,
    connection: SnowflakeConnection | None = None,
):
    client = Client(Resources(profile, config, connection), rel.Compiler(config), name, dry_run=dry_run)
    client.exec_control("def insert[:rel, :config]: :disable_ivm")
    client.install("pyrel_base", dsl.build.raw_task("""
        @inline
        def make_identity(x..., z):
            exists((u) | hash128({x...}, x..., u) and
            hash_value_uint128_convert(u, z))

        @inline
        def pyrel_default({F}, c, k..., v):
            F(k..., v) or (not F(k..., _) and v = c)

        @inline
        def pyrel_unwrap(x in UInt128, y): y = x
    """))
    return dsl.Graph(client, name)



from pathlib import Path
import random
import os

import pytest
import snowflake.connector

import relationalai as rai
from relationalai import debugging
from relationalai.debugging import logger
from relationalai.clients import config as cfg
from relationalai.util.snowflake_logger import SnowflakeLogger
from testutil.snapshot import path_to_slug, validate_block_snapshots

def engine_config_fixture(size):
    # Check for an externally provided engine name
    # It is used in GitHub Actions to run tests against a specific engine
    engine_name = os.getenv("ENGINE_NAME")
    if engine_name:
        # If engine name was provided, just yield the config
        config = make_config(engine_name)
        yield config
        return

    # If there's a local config file, use it, including
    # the engine specified there.
    config = cfg.Config()
    if config.file_path is not None:
        yield config
        return

    # Otherwise, create a new engine and delete it afterwards.
    random_number = random.randint(1000000000, 9999999999)
    engine_name = f"pyrel_test_{random_number}"
    create_engine(engine_name, size=size)

    yield make_config(engine_name)

    delete_engine(engine_name)

def create_engine(engine_name: str, size: str):
    config = make_config(engine_name)

    sf_compute_pool = os.getenv("SF_TEST_COMPUTE_POOL", config.get("compute_pool", ""))

    provider = rai.Resources(config=config)
    print(f"Creating engine {engine_name}")
    provider.create_engine(name=engine_name, size=size, pool=sf_compute_pool)
    print(f"Engine {engine_name} created")


def delete_engine(engine_name: str):
    print(f"Deleting engine {engine_name}")
    config = make_config(engine_name)
    provider = rai.Resources(config=config)
    provider.delete_engine(engine_name)
    print(f"Engine {engine_name} deleted")

def make_config(engine_name: str) -> cfg.Config:
    cloud_provider = os.getenv("RAI_CLOUD_PROVIDER")

    match cloud_provider:
        case None:
            raise ValueError("RAI_CLOUD_PROVIDER must be set")
        case "azure":
            client_id = os.getenv("RAI_CLIENT_ID")
            client_secret = os.getenv("RAI_CLIENT_SECRET")
            if client_id is None or client_secret is None:
                raise ValueError(
                    "RAI_CLIENT_ID, RAI_CLIENT_SECRET must be set if RAI_CLOUD_PROVIDER is set to 'azure'"
                )
            # Running against prod
            return cfg.Config(
                {
                    "platform": "azure",
                    "host": "azure.relationalai.com",
                    "port": "443",
                    "region": "us-east",
                    "scheme": "https",
                    "client_credentials_url": "https://login.relationalai.com/oauth/token",
                    "client_id": client_id,
                    "client_secret": client_secret,
                    "engine": engine_name,
                }
            )
        case "snowflake":
            sf_username = os.getenv("SF_TEST_ACCOUNT_USERNAME")
            sf_password = os.getenv("SF_TEST_ACCOUNT_PASSWORD")
            sf_account = os.getenv("SF_TEST_ACCOUNT_NAME")
            sf_warehouse = os.getenv("SF_TEST_WAREHOUSE_NAME")
            sf_app_name = os.getenv("SF_TEST_APP_NAME")
            sf_compute_pool = os.getenv("SF_TEST_COMPUTE_POOL")
            if sf_username is None or sf_password is None:
                raise ValueError(
                    "SF_TEST_ACCOUNT_USERNAME, SF_TEST_ACCOUNT_PASSWORD, SF_TEST_ACCOUNT_NAME must be set if RAI_CLOUD_PROVIDER is set to 'snowflake'"
                )
            return cfg.Config(
                {
                    "platform": "snowflake",
                    "user": sf_username,
                    "password": sf_password,
                    "account": sf_account,
                    "role": "rai_integration_consumer",
                    "warehouse": sf_warehouse,
                    "rai_app_name": sf_app_name,
                    "engine": engine_name,
                    "compute_pool": sf_compute_pool,
                }
            )
        case _:
            raise ValueError(f"Unsupported cloud provider: {cloud_provider}")

def snowflake_logger_fixture():
    if not os.getenv("SF_REPORTING_PASSWORD"):
        print('benchmarking logger disabled since config env vars not present')
        yield
        return

    conn = snowflake.connector.connect(
        user=os.getenv("SF_REPORTING_USER"),
        password=os.getenv("SF_REPORTING_PASSWORD"),
        account=os.getenv("SF_REPORTING_ACCOUNT"),
        role=os.getenv("SF_REPORTING_ROLE"),
        warehouse=os.getenv("SF_REPORTING_WAREHOUSE"),
        database=os.getenv("SF_REPORTING_DATABASE"),
        schema=os.getenv("SF_REPORTING_SCHEMA"),
    )
    snowflake_logger = SnowflakeLogger(conn)
    logger.addHandler(snowflake_logger)
    yield
    snowflake_logger.shut_down()

def root_span_fixture():
    git_info = get_git_info()
    attrs = {
        **git_info,
        "cloud_provider": os.getenv("RAI_CLOUD_PROVIDER"),
    }
    with debugging.span("test_session", **attrs):
        yield

def get_git_info():
    if os.getenv("GITHUB_REF_NAME") is not None:
        return {
            'commit': os.getenv("GITHUB_SHA"),
            'branch': os.getenv("GITHUB_REF_NAME"),
        }
    else:
        # shell out to get the info
        return {
            'commit': 'local',
            'branch': 'local',
        }

def randomize(name: str) -> str:
    return f"{name}_{random.randint(1000000000, 9999999999)}"

def do_snapshot_test(config: cfg.Config, engine_size: str, snapshot, file_path: Path):
    """
    Run the PyRel program in the given `file_path`, testing the results against a snapshot file,
    using the given `config` and `engine_size`.
    """
    
    test_name = file_path.stem

    provider = rai.Resources(config=config)
    
    def get_results(event):
        if "results" in event:
            return str(event["results"]["values"])
    
    with debugging.span("test", name=test_name, engine_size=engine_size):
        for i in range(1):
            with debugging.span("run", idx=i):
                db_name = randomize(test_name)
                validate_block_snapshots(file_path, snapshot, get_results, "query", {
                    "name": db_name,
                    "config": config,
                })
                try:
                    provider.delete_graph(db_name)
                except Exception as e:
                    print(f"Failed to delete graph {db_name}: {e}")

def all_files_in_dir(test_case_dir: Path):
    test_case_files = [path for path in test_case_dir.rglob("*.py")]
    print('files:', test_case_files)
    return pytest.mark.parametrize(
        "file_path", test_case_files, ids=lambda path: path_to_slug(path, test_case_dir)
    )



from collections import defaultdict
from copy import deepcopy
from numbers import Number
import textwrap
from typing import Any
from .. import dsl, errors
import gravis as gv
from . import rel, Vars

#--------------------------------------------------
# Errors
#--------------------------------------------------

def invalid_param(param:str, message:str):
    source = errors.Errors.call_source(4)
    marked = errors.mark_source(source, source.line, source.line) if source else ""
    content = f"""
    The [yellow]{param}[/yellow] parameter {message}

    {marked}
    """
    errors.print_source_error(source, "Invalid algorithm parameter", content)
    raise ValueError(message)

#--------------------------------------------------
# Validation
#--------------------------------------------------

def between(param:str,value, min, max,  type_=Number):
    if (not isinstance(value, type_)) or (value < min or value > max):
        invalid_param(param, f"must be a {type_.__name__.lower()} between {min} and {max}, got {value!r}")

def positive(param:str, value, type_=Number):
    if (not isinstance(value, type_)) or (value <= 0):
        invalid_param(param, f"must be a positive {type_.__name__.lower()}, got {value!r}")

def nonnegative(param:str, value, type_=Number):
    if (not isinstance(value, type_)) or (value < 0):
        invalid_param(param, f"must be a non-negative {type_.__name__.lower()}, got {value!r}")

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def unwrap(v):
    if v is None:
        return None
    return rel.pyrel_default(rel.pyrel_unwrap, v, v)


#--------------------------------------------------
# Algos
#--------------------------------------------------

class Compute:
    def __init__(self, graph:'Graph'):
        self._graph = graph
        self._lib = dsl.global_ns.graphlib

    def _config(self, **kwargs):
        return dsl.InlineRelation(self._graph.model, [
            *[(dsl.Symbol(k), v) for k, v in kwargs.items()],
        ])

    def _lookup(self, v):
        res = Vars(1)
        lookup_rel = getattr(rel, f"{self._graph._graph_ref()}_lookup")
        rel.pyrel_default(lookup_rel, v, v, res)
        return res

    # --------------------------------------------------
    # Degree
    # --------------------------------------------------

    def degree(self, node, weight=None):
        if not weight:
            return self._lib.degree(self._graph, unwrap(node))
        return self._lib.weighted_degree(self._graph, unwrap(node))

    def min_degree(self):
        return self._lib.min_degree(self._graph)

    def max_degree(self):
        return self._lib.max_degree(self._graph)

    def avg_degree(self):
        return self._lib.average_degree(self._graph)

    def indegree(self, node):
        return self._lib.indegree(self._graph, unwrap(node))

    def min_indegree(self):
        return self._lib.min_indegree(self._graph)

    def max_indegree(self):
        return self._lib.max_indegree(self._graph)

    def avg_indegree(self):
        return self._lib.average_indegree(self._graph)

    def outdegree(self, node):
        return self._lib.outdegree(self._graph, unwrap(node))

    def min_outdegree(self):
        return self._lib.min_outdegree(self._graph)

    def max_outdegree(self):
        return self._lib.max_outdegree(self._graph)

    def avg_outdegree(self):
        return self._lib.average_outdegree(self._graph)

    # --------------------------------------------------
    # Basics
    # --------------------------------------------------

    def num_nodes(self):
        return self._lib.num_nodes(self._graph)

    def num_edges(self):
        return self._lib.num_edges(self._graph)

    # --------------------------------------------------
    # Distance
    # --------------------------------------------------

    def distance(self, node1, node2):
        return self._lib.distance(self._graph, unwrap(node1), unwrap(node2))

    def weighted_distance(self, node1, node2):
        return self._lib.weighted_distance(self._graph, unwrap(node1), unwrap(node2))

    def diameter_range(self):
        min, max = Vars(2)
        self._lib.diameter_range(self._graph, "min", min)
        self._lib.diameter_range(self._graph, "max", max)
        return (min, max)

    # --------------------------------------------------
    # Connectivity
    # --------------------------------------------------

    def is_connected(self):
        model = self._graph.model
        with model.ordered_choice() as v:
            with model.scope():
                self._lib.is_connected(self._graph)
                v.add(True)
            with model.scope():
                v.add(False)
        return v

    def reachable_from(self, node: dsl.Producer) -> dsl.Expression:
        return self._lib.transitive_closure(self._graph, unwrap(node))

    def is_reachable(self, node1: dsl.Producer, node2: dsl.Producer) -> dsl.Expression:
        model = self._graph.model
        with model.ordered_choice() as v:
            with model.scope():
                self._lib.transitive_closure(self._graph, unwrap(node1), unwrap(node2))
                v.add(True)
            with model.scope():
                v.add(False)
        return v

    # --------------------------------------------------
    # Triangles
    # --------------------------------------------------

    def num_triangles(self, node=None):
        if node is None:
            return self._lib.num_triangles(self._graph)
        return self._lib.triangle_count(self._graph, unwrap(node))

    def is_triangle(self, node1, node2, node3):
        model = dsl.get_graph()
        with model.ordered_choice() as v:
            with model.scope():
                self._lib.triangle(self._graph, unwrap(node1), unwrap(node2), unwrap(node3))
                v.add(True)
            with model.scope():
                v.add(False)
        return v

    def triangles(self, node=None):
        # If no node is provided, return all triangles in the graph,
        # unique up to ordering of the nodes.
        if node is None:
            x, y, z = Vars(3)
            self._lib.unique_triangle(self._graph, x, y, z)
            lx = self._lookup(x)
            ly = self._lookup(y)
            lz = self._lookup(z)
            return (lx, ly, lz)
        # If a node is provided, return all triangles that include that node,
        # unique up to order of the nodes.
        x, y = Vars(2)
        model = dsl.get_graph()
        node = unwrap(node)
        with model.union(dynamic=True) as t:
            for (a, b, c) in ([node, x, y], [x, node, y], [x, y, node]):
                with model.scope():
                    self._lib.unique_triangle(self._graph, a, b, c)
                    t.add(node, a=a, b=b, c=c)
        la = self._lookup(t.a)
        lb = self._lookup(t.b)
        lc = self._lookup(t.c)
        return (la, lb, lc)

    # --------------------------------------------------
    # Clustering
    # --------------------------------------------------

    def local_clustering_coefficient(self, node):
        if not self._graph.undirected:
            invalid_param("graph", "must be undirected")
        return self._lib.local_clustering_coefficient(self._graph, unwrap(node))

    def avg_clustering_coefficient(self):
        if not self._graph.undirected:
            invalid_param("graph", "must be undirected")
        return self._lib.average_clustering_coefficient(self._graph)

    # --------------------------------------------------
    # Link Prediction
    # --------------------------------------------------

    def adamic_adar(self, node1, node2):
        return self._lib.adamic_adar(self._graph, unwrap(node1), unwrap(node2))

    def preferential_attachment(self, node1, node2):
        return self._lib.preferential_attachment(self._graph, unwrap(node1), unwrap(node2))

    def common_neighbor(self, node1, node2):
        return self._lib.common_neighbor(self._graph, unwrap(node1), unwrap(node2))

    # --------------------------------------------------
    # Similarity
    # --------------------------------------------------

    def cosine_similarity(self, node1, node2):
        if not self._graph.undirected:
            invalid_param("graph", "must be undirected")
        return self._lib.cosine_similarity(self._graph, unwrap(node1), unwrap(node2))

    def weighted_cosine_similarity(self, node1, node2):
        if not self._graph.undirected:
            invalid_param("graph", "must be undirected")
        return self._lib.weighted_cosine_similarity(self._graph, unwrap(node1), unwrap(node2))

    def jaccard_similarity(self, node1, node2):
        return self._lib.jaccard_similarity(self._graph, unwrap(node1), unwrap(node2))

    def weighted_jaccard_similarity(self, node1, node2):
        return self._lib.weighted_jaccard_similarity(self._graph, unwrap(node1), unwrap(node2))

    # --------------------------------------------------
    # Centrality
    # --------------------------------------------------

    def pagerank(self, node, damping_factor=0.85, tolerance=1e-6, max_iter=20):
        between("damping_factor", damping_factor, 0, 1)
        positive("tolerance", tolerance)
        positive("max_iter", max_iter, type_=int)

        config = self._config(
            graph=self._graph,
            damping_factor=float(damping_factor),
            tolerance=float(tolerance),
            max_iter=max_iter,
        )
        return self._lib.pagerank(config, unwrap(node))

    def betweenness_centrality(self, node):
        return self._lib.betweenness_centrality(self._graph, unwrap(node))

    def degree_centrality(self, node):
        return self._lib.degree_centrality(self._graph, unwrap(node))

    def eigenvector_centrality(self, node):
        if not self._graph.undirected:
            invalid_param("graph", "must be undirected")
        return self._lib.eigenvector_centrality(self._graph, unwrap(node))

    def weighted_degree_centrality(self, node):
        return self._lib.weighted_degree_centrality(self._graph, unwrap(node))

    # --------------------------------------------------
    # Community Detection
    # --------------------------------------------------

    def weakly_connected_component(self, node):
        component = self._lib.weakly_connected_component(self._graph, unwrap(node))
        return self._lookup(component)

    def triangle_community(self, node):
        return self._lib.triangle_community(self._graph, unwrap(node))

    def infomap(
            self,
            node,
            max_levels: int = 4,
            max_sweeps: int = 8,
            level_tolerance: float = 0.01,
            sweep_tolerance: float = 0.0001,
            teleportation_rate: float = 0.15,
            visit_rate_tolerance: float = 1e-15,
            randomization_seed: int | None = None,
    ):
        positive("max_levels", max_levels, type_=int)
        nonnegative("max_sweeps", max_sweeps, type_=int)
        nonnegative("level_tolerance", level_tolerance)
        nonnegative("sweep_tolerance", sweep_tolerance)
        positive("teleportation_rate", teleportation_rate)
        between("teleportation_rate", teleportation_rate, 0, 1)
        positive("visit_rate_tolerance", visit_rate_tolerance)
        _config_dict = {
            "graph": self._graph,
            "max_levels": max_levels,
            "max_sweeps": max_sweeps,
            "level_tolerance": float(level_tolerance),
            "sweep_tolerance": float(sweep_tolerance),
            "teleportation_rate": float(teleportation_rate),
            "visit_rate_tolerance": float(visit_rate_tolerance),
        }
        if randomization_seed is not None:
            positive("randomization_seed", randomization_seed, type_=int)
            _config_dict["randomization_seed"] = randomization_seed

        config = self._config(**_config_dict)
        return self._lib.infomap(config, unwrap(node))

    def louvain(
            self,
            node,
            max_levels: int = 4,
            max_sweeps: int = 8,
            level_tolerance: float = 0.01,
            sweep_tolerance: float = 0.0001,
            randomization_seed: int | None = None,
    ):
        positive("max_levels", max_levels, type_=int)
        nonnegative("max_sweeps", max_sweeps, type_=int)
        nonnegative("level_tolerance", level_tolerance)
        nonnegative("sweep_tolerance", sweep_tolerance)
        _config_dict = {
            "graph": self._graph,
            "max_levels": max_levels,
            "max_sweeps": max_sweeps,
            "level_tolerance": float(level_tolerance),
            "sweep_tolerance": float(sweep_tolerance),
        }
        if randomization_seed is not None:
            positive("randomization_seed", randomization_seed, type_=int)
            _config_dict["randomization_seed"] = randomization_seed
        config = self._config(**_config_dict)
        return self._lib.louvain(config, unwrap(node))

    def label_propagation(self, node, max_sweeps: int = 1000, randomization_seed: int | None = None):
        positive("max_sweeps", max_sweeps, type_=int)
        _config_dict = {
            "graph": self._graph,
            "max_sweeps": max_sweeps,
        }
        if randomization_seed is not None:
            positive("randomization_seed", randomization_seed, type_=int)
            _config_dict["randomization_seed"] = randomization_seed
        config = self._config(**_config_dict)
        return self._lib.label_propagation(config, unwrap(node))

#--------------------------------------------------
# Edge
#--------------------------------------------------

class EdgeInstance:
    def __init__(self, edge:'Edge', from_:Any, to:Any, kwargs:dict={}):
        self._edge = edge
        self._graph = edge._graph
        self.from_ = from_
        self.to = to
        for k, v in kwargs.items():
            self._edge._prop(k)(from_, to, v)

    def __getattr__(self, name:str):
        v = Vars(1)
        self._edge._prop(name)(self.from_, self.to, v)
        return v

    def set(self, **kwargs):
        for k, v in kwargs.items():
            self._edge._prop(k).add(self.from_, self.to, v)
        return self

    def _to_var(self):
        raise Exception("Edges can't be returned directly, you can return the from_ and to properties individually")

class Edge:
    def __init__(self, graph:'Graph'):
        self._graph = graph
        self._type = dsl.RawRelation(self._graph.model, f"graph{self._graph.id}_edges", 2)
        self._props = {}

    def _prop(self, name:str):
        if name not in self._props:
            self._props[name] = dsl.RawRelation(self._graph.model, f"graph{self._graph.id}_edge_{name}", 3)
        return self._props[name]

    def _is_weighted(self):
        return "weight" in self._props

    def extend(self, prop:'dsl.Property', **kwargs):
        type = prop._provider
        with self._graph.model.rule():
            t = type()
            self.add(t, getattr(t, prop._name), **kwargs)

    def add(self, from_:Any, to:Any, **kwargs):
        self._type.add(from_, to)
        for k, v in kwargs.items():
            self._prop(k).add(from_, to, v)
        return EdgeInstance(self, from_, to, {})

    def __call__(self, from_:Any=None, to:Any=None, **kwargs):
        if from_ is None:
            from_ = Vars(1)
        if to is None:
            to = Vars(1)
        self._type(from_, to)
        return EdgeInstance(self, from_, to, kwargs)

    def __getattribute__(self, __name: str) -> Any:
        if __name in ["add", "extend"] or __name.startswith("_"):
            return super().__getattribute__(__name)
        return self._props[__name]

#--------------------------------------------------
# Graph
#--------------------------------------------------

class Graph:
    def __init__(self, model:dsl.Graph, undirected=False, weighted=False, default_weight=1.0, with_isolated_nodes=False):
        self.model = model
        self.id = dsl.next_id()
        self.compute = Compute(self)
        self.Node = dsl.Type(model, "nodes", scope=f"graph{self.id}_")
        self.Edge = Edge(self)
        self.undirected = undirected
        self.weighted = weighted
        self.default_weight=default_weight
        self._last_fetch = None

        graph_type = "::graphlib::undirected_graph" if undirected else "::graphlib::directed_graph"
        if weighted:
            unwrapped = f"""
            def {self.Edge._prop("weight")._name}_unwrapped(au, bu, w): {{
                exists((a, b) |
                    {self.Edge._prop("weight")._name}(a, b, w) and
                    pyrel_default(pyrel_unwrap, a, a, au) and
                    pyrel_default(pyrel_unwrap, b, b, bu)
                )
            }}
            """
            create_graph = f"""{graph_type}[{self.Edge._prop("weight")._name}_unwrapped]"""
        else:
            unwrapped = f"""
            def {self.Edge._type._name}_unwrapped(au, bu): {{
                exists((a, b) |
                    {self.Edge._type._name}(a, b) and
                    pyrel_default(pyrel_unwrap, a, a, au) and
                    pyrel_default(pyrel_unwrap, b, b, bu)
                )
            }}
            """
            create_graph = f"{graph_type}[{self.Edge._type._name}_unwrapped]"

        node_def = ""
        if with_isolated_nodes:
            node_def = f"def {self._graph_ref()}[:node]: {{pyrel_unwrap[{self.Node._type.name}]}}"

        self.model.install_raw(textwrap.dedent(f"""
        declare {self.Node._type.name}
        declare {self.Edge._type._name}
        {unwrapped}
        @inline
        def {self._graph_ref()}_lookup(uw, orig): {{ {self.Node._type.name}(orig) and pyrel_unwrap(orig, uw) }}

        {f"declare {self.Edge._prop('weight')._name}" if weighted else ""}
        def {self._graph_ref()} {{{create_graph}}}
        {node_def}
        """))

        # Add a rule that makes all nodes used in edges are also added to
        # the nodes relation
        with model.rule():
            a, b = dsl.Vars(2)
            self.Edge._type(a, b)
            self.Node.add(a)
            self.Node.add(b)

    def _graph_ref(self):
        return f"graph{self.id}"

    def _to_var(self):
        return getattr(rel, self._graph_ref())._to_var()

    def _is_weighted(self):
        return self.weighted
        # return self.edges._is_weighted()

    #--------------------------------------------------
    # Fetch
    #--------------------------------------------------

    def fetch(self):
        code = []
        code.append(f"def output(:nodes, n): {self.Node._type.name}(n)")
        for prop in self.Node._type.properties:
            scope = self.Node._scope
            code.append(f"def output(:nodes, :{prop.name.removeprefix(scope)}, n, v): {prop.name}(n, v)")
        code.append(f"def output(:edges, a,b): {self.Edge._type._name}(a,b)")
        for name, prop in self.Edge._props.items():
            code.append(f"def output(:edges, :{name}, a, b, v): {prop._name}(a, b, v)")

        output = {"nodes": defaultdict(dict), "edges": defaultdict(dict)}
        results = self.model.exec_raw("\n".join(code), raw_results=True)
        for set_ in results.results:
            path = [v[1:] for v in set_["relationId"].split("/")[2:] if v[0] == ":"]
            if path[0] not in ["nodes", "edges"]:
                continue
            cur = output[path[0]]
            if path[0] == "nodes":
                if len(path) == 1:
                    for (n,) in set_["table"].itertuples(index=False):
                        cur[n]
                else:
                    for (n, v) in set_["table"].itertuples(index=False):
                        cur[n][path[1]] = v
            elif path[0] == "edges":
                if len(path) == 1:
                    for (a, b) in set_["table"].itertuples(index=False):
                        # If the graph is undirected, normalize the order of
                        # the source and target nodes.
                        if self.undirected and a > b:
                            a, b = b, a
                        cur[(a, b)]
                else:
                    for (a, b, v) in set_["table"].itertuples(index=False):
                        if self.undirected and a > b:
                            a, b = b, a
                        cur[(a, b)][path[1]] = v
            else:
                raise Exception(f"Unexpected path: {path}")
        self._last_fetch = output
        return output

    #--------------------------------------------------
    # Visualize
    #--------------------------------------------------

    default_visual_props = {
        "node": {
            "color": "black",
            "opacity": 1,
            "size": 10,
            "shape": "circle",
            "border_color": "black",
            "border_size": 1,
            "label_color": "black",
            "label_size": 10,
        },
        "edge": {
            "color": "#999",
            "opacity": 1,
            "size": 2,
            "shape": "circle",
            "border_color": "#999",
            "border_size": 1,
            "label_color": "black",
            "label_size": 10,
            "arrow_size": 4,
            "arrow_color": "#999",
        }
    }

    _style_map = {
        "arrow_size": "arrow_size",
        "arrow_color": "arrow_color",
    }

    def _visual_props(self, prop_def, metadata):
        for k, v in prop_def.items():
            if callable(v):
                metadata[k] = v(metadata)
        return metadata

    def _visual_dict(self, style: dict, use_cache = False) -> dict:
        data = self._last_fetch if use_cache else None
        if not data:
            data = self.fetch()
        style = deepcopy(style)
        if "node" not in style:
            style["node"] = {}
        if "edge" not in style:
            style["edge"] = {}
        merged_style = deepcopy(self.default_visual_props)
        for category in ["node", "edge"]:
            for k, v in style.get(category, {}).items():
                if not callable(v):
                    merged_style[category][k] = v
        flat_style = {
            self._style_map.get(k, f"{type}_{k}"): v
            for type, category in merged_style.items()
            for k, v in category.items()
        }

        return {
            "graph": {
                "directed": not self.undirected,
                "metadata": flat_style,
                "nodes": {
                    node_id: {
                        **({"label": info["label"]} if "label" in info else {}),
                        "metadata": self._visual_props(style["node"], info),
                    }
                    for (node_id, info) in data["nodes"].items()
                },
                "edges": [
                    {"source": source, "target": target, "metadata": self._visual_props(style["edge"], info)}
                    for ((source, target), info) in data["edges"].items()
                ],
            }
        }

    def visualize(self, three=False, style: dict | None = None, use_cache = False, **kwargs):
        style = style if style is not None else {"node": {}, "edge": {}}
        vis = gv.vis if not three else gv.three
        graph_dict = self._visual_dict(style=style, use_cache=use_cache)
        # Use defaults for the following kwargs if not provided
        new_kwargs = {
            "node_label_data_source": "label",
            "edge_label_data_source": "label",
            "show_edge_label": True,
            "edge_curvature": 0.4,
        } | kwargs
        fig = vis(graph_dict, **new_kwargs)
        return fig

#--------------------------------------------------
# Path
#--------------------------------------------------

class Path:
    def __init__(self, *args):
        self.edges = []
        self.nodes = []
        pass

    def __getitem__(self, item):
        return self

#--------------------------------------------------
# Exports
#--------------------------------------------------

__all__ = ["Graph", "Path"]
